"""
Created on Fri Sep 20 14:55:43 2024

@author: mcastano
Multi-Fidelity co-Kriging: recursive formulation with autoregressive model of
order 1 (AR1)
Adapted on January 2021 by Andres Lopez-Lopera to the new SMT version
"""

import warnings
from copy import deepcopy

import numpy as np
from scipy import linalg
from scipy.cluster.vq import kmeans
from scipy.linalg import solve_triangular
from sklearn.cross_decomposition import PLSRegression as pls

from smt.surrogate_models.krg_based import (
    KrgBased,
    MixIntKernelType,
    compute_n_param,
)
from smt.surrogate_models.sgp import SGP
from smt.utils.kriging import (
    componentwise_distance,
    compute_X_cont,
    cross_distances,
    differences,
)
from smt.utils.misc import standardization


class SMFK(KrgBased):
    def _initialize(self):
        super()._initialize()
        declare = self.options.declare
        declare(
            "rho_regr",
            "constant",
            values=("constant", "linear", "quadratic"),
            desc="Regression function type for rho",
        )
        declare(
            "optim_var",
            False,
            types=bool,
            values=(True, False),
            desc="If True, the variance at HF samples is forced to zero",
        )
        declare(
            "n_inducing",
            4,
            types=(int),
            desc="Number of inducing points for the lowest fidelity \
                level must be less or equal to the DoE in the lowest fideliy.",
        )
        declare(
            "propagate_uncertainty",
            True,
            types=bool,
            values=(True, False),
            desc="If True, the variance cotribution of lower fidelity levels are considered",
        )
        self.name = "SMFK"

        self.X2_norma = {}
        self.X2_offset = {}
        self.X2_scale = {}
        self.Z = None

    def _differences(self, X, Y):
        """
        Compute the distances
        """
        return differences(X, Y)

    def _check_list_structure(self, X, y):
        """
        checks if the data structure is compatible with MFK.
        sets class attributes such as (number of levels of Fidelity, training points in each level, ...)

        Arguments :
        X : list of arrays, each array corresponds to a fidelity level. starts from lowest to highest
        y : same as X
        """

        if not isinstance(X, list):
            nlevel = 1
            X = [X]
        else:
            nlevel = len(X)

        if not isinstance(y, list):
            y = [y]

        if len(X) != len(y):
            raise ValueError("X and y must have the same length.")

        n_samples = np.zeros(nlevel, dtype=int)
        n_features = np.zeros(nlevel, dtype=int)
        n_samples_y = np.zeros(nlevel, dtype=int)
        for i in range(nlevel):
            n_samples[i], n_features[i] = X[i].shape
            if i > 1 and n_features[i] != n_features[i - 1]:
                raise ValueError("All X must have the same number of columns.")
            y[i] = np.asarray(y[i]).ravel()[:, np.newaxis]
            n_samples_y[i] = y[i].shape[0]
            if n_samples[i] != n_samples_y[i]:
                raise ValueError("X and y must have the same number of rows.")

        self.nx = n_features[0]
        self.nt_all = n_samples
        self.nlvl = nlevel
        self.ny = y[0].shape[1]
        self.X = X[:]
        self.y = y[:]

    def _new_train(self):
        """
        Overrides KrgBased implementation
        Trains the Multi-Fidelity model
        """
        self._corr_params = None

        self._new_train_init()
        theta0 = self.options["theta0"].copy()
        noise0 = self.options["noise0"].copy()

        for lvl in range(self.nlvl):
            print("Optimizing Level:", lvl)
            self._new_train_iteration(lvl)
            self.options["theta0"] = theta0
            self.options["noise0"] = noise0

        self._reinterpolate(lvl)

    def _new_train_init(self):
        if self.name in ["MFKPLS", "MFKPLSK"]:
            _pls = pls(self.options["n_comp"])

            # As of sklearn 0.24.1 PLS with zeroed outputs raises an exception
            # while sklearn 0.23 returns zeroed x_rotations
            # For now the try/except below is a workaround to restore the 0.23 behaviour
            try:
                # PLS is done on the highest fidelity identified by the key None
                self.m_pls = _pls.fit(
                    self.training_points[None][0][0].copy(),
                    self.training_points[None][0][1].copy(),
                )
                self.coeff_pls = self.m_pls.x_rotations_
            except StopIteration:
                self.coeff_pls = np.zeros(
                    self.training_points[None][0][0].shape[1], self.options["n_comp"]
                )

        xt = []
        yt = []
        i = 0
        while self.training_points.get(i, None) is not None:
            xt.append(self.training_points[i][0][0])
            yt.append(self.training_points[i][0][1])
            i = i + 1
        xt.append(self.training_points[None][0][0])
        yt.append(self.training_points[None][0][1])

        self._check_list_structure(xt, yt)
        self._check_param()
        X = self.X
        y = self.y

        if self.is_continuous:
            (
                _,
                _,
                self.X_offset,
                self.y_mean,
                self.X_scale,
                self.y_std,
            ) = standardization(np.concatenate(xt, axis=0), np.concatenate(yt, axis=0))
            self.cat_features = [False] * np.shape(np.concatenate(xt, axis=0))[1]
        else:
            (
                _,
                _,
                self.X_offset,
                self.y_mean,
                self.X_scale,
                self.y_std,
            ) = standardization(np.concatenate(xt, axis=0), np.concatenate(yt, axis=0))
            _, self.cat_features = compute_X_cont(
                np.concatenate(xt, axis=0), self.design_space
            )
            self.X_offset[self.cat_features] *= 0
            self.X_scale[self.cat_features] *= 0
            self.X_scale[self.cat_features] += 1

        nlevel = self.nlvl

        # initialize lists
        self.optimal_noise_all = nlevel * [0]
        self.D_all = nlevel * [0]
        self.F_all = nlevel * [0]
        self.p_all = nlevel * [0]
        self.q_all = nlevel * [0]
        self.optimal_rlf_value = nlevel * [0]
        self.optimal_par = nlevel * [{}]
        self.optimal_theta = nlevel * [0]
        self.X_norma_all = [(x - self.X_offset) / self.X_scale for x in X]
        self.y_norma_all = [(f - self.y_mean) / self.y_std for f in y]

        self.X_norma_all = [(x - self.X_offset) / self.X_scale for x in X]

        self.y_norma_all = [(f - self.y_mean) / self.y_std for f in y]

        if not (self.is_continuous):
            x2t, _ = self.design_space.unfold_x(np.concatenate(xt, axis=0))
            _, _, self.X2_offset, _, self.X2_scale, _ = standardization(
                x2t, np.concatenate(yt, axis=0)
            )

    def _new_train_iteration(self, lvl):
        # n_samples = self.nt_all
        self.options["noise0"] = np.array([self.options["noise0"][lvl]]).flatten()
        self.options["theta0"] = self.options["theta0"][lvl, :]

        self.X_norma = self.X_norma_all[lvl]
        self.y_norma = self.y_norma_all[lvl]

        if self.options["eval_noise"]:
            if self.options["use_het_noise"]:
                # hetGP works with unique design variables
                (
                    self.X_norma,
                    self.index_unique,  # do we need to store it?
                    self.nt_reps,  # do we need to store it?
                ) = np.unique(
                    self.X_norma, return_inverse=True, return_counts=True, axis=0
                )
                self.nt_all[lvl] = self.X_norma.shape[0]

                # computing the mean of the output per unique design variable (see Binois et al., 2018)
                y_norma_unique = []
                for i in range(self.nt_all[lvl]):
                    y_norma_unique.append(np.mean(self.y_norma[self.index_unique == i]))
                y_norma_unique = np.array(y_norma_unique).reshape(-1, 1)

                # pointwise sensible estimates of the noise variances (see Ankenman et al., 2010)
                self.optimal_noise = self.options["noise0"] * np.ones(self.nt_all[lvl])
                for i in range(self.nt_all[lvl]):
                    diff = self.y_norma[self.index_unique == i] - y_norma_unique[i]
                    if np.sum(diff**2) != 0.0:
                        self.optimal_noise[i] = np.std(diff, ddof=1) ** 2
                self.optimal_noise = self.optimal_noise / self.nt_reps
                self.optimal_noise_all[lvl] = self.optimal_noise
                self.y_norma = y_norma_unique

                self.X_norma_all[lvl] = self.X_norma
                self.y_norma_all[lvl] = self.y_norma
        else:
            self.optimal_noise = self.options["noise0"] / self.y_std**2
            self.optimal_noise_all[lvl] = self.optimal_noise

        # Calculate matrix of distances D between samples
        if self.is_continuous:
            if lvl != 0:
                self.D_all[lvl] = cross_distances(self.X_norma)

        # Regression matrix and parameters
        if lvl != 0:
            self.F_all[lvl] = self._regression_types[self.options["poly"]](self.X_norma)
            self.p_all[lvl] = self.F_all[lvl].shape[1]

        if lvl == 0:
            # l = np.std(self.X_norma, axis=0)
            # theta = 1/l**2
            # bounds = [1e-8, 1e2]

            data = np.hstack((self.X_norma, self.y_norma))
            Z2 = kmeans(data, self.options["n_inducing"])[0][:, :-1]

            sgp = SGP(
                method="FITC",
                nugget=5000 * np.finfo(np.double).eps,
                print_prediction=False,
                print_global=False,
            )
            sgp.set_training_values(self.X_norma, self.y_norma)
            sgp.set_inducing_inputs(Z=Z2)
            sgp.train()

            self.sgp = sgp
            self.Z = Z2 * self.X_scale + self.X_offset

        # Concatenate the autoregressive part for levels > 0
        if lvl > 0:
            D, self.ij = self.D_all[lvl]
            self.F = self.F_all[lvl]
            self._lvl = lvl
            self.nt = self.nt_all[lvl]
            self.q = self.q_all[lvl]
            self.p = self.p_all[lvl]
            self.kplsk_second_loop = False

            F_rho = self._regression_types[self.options["rho_regr"]](self.X_norma)
            self.q_all[lvl] = F_rho.shape[1]

            if self.is_continuous:
                if lvl == 1:
                    pred_old = self.sgp.predict_values(self.X_norma)
                else:
                    pred_old = self._predict_intermediate_values(
                        self.X_norma, lvl, descale=False
                    )

                self.F_all[lvl] = np.hstack(
                    (
                        F_rho
                        * np.dot(
                            pred_old,
                            np.ones((1, self.q_all[lvl])),
                        ),
                        self.F_all[lvl],
                    )
                )
            else:
                self.F_all[lvl] = np.hstack(
                    (
                        F_rho
                        * np.dot(
                            self._predict_intermediate_values(
                                self.X[lvl], lvl, descale=False
                            ),
                            np.ones((1, self.q_all[lvl])),
                        ),
                        self.F_all[lvl],
                    )
                )
            # n_samples_F_i = self.F_all[lvl].shape[0]

            self.F = self.F_all[lvl]
            D, self.ij = self.D_all[lvl]
            self._lvl = lvl
            self.nt = self.nt_all[lvl]
            self.q = self.q_all[lvl]
            self.p = self.p_all[lvl]
            self.kplsk_second_loop = False
            (
                self.optimal_rlf_value[lvl],
                self.optimal_par[lvl],
                self.optimal_theta[lvl],
            ) = self._optimize_hyperparam(D)

            if self.options["eval_noise"] and not self.options["use_het_noise"]:
                tmp_list = self.optimal_theta[lvl]
                self.optimal_theta[lvl] = tmp_list[:-1]
                self.optimal_noise = tmp_list[-1]
                self.optimal_noise_all[lvl] = self.optimal_noise
            del self.y_norma, self.D, self.optimal_noise

        # if n_samples_F_i != n_samples[lvl]:
        #     raise Exception(
        #         "Number of rows in F and X do not match. Most "
        #         "likely something is going wrong with the "
        #         "regression model."
        #     )

        # if int(self.p_all[lvl] + self.q_all[lvl]) >= n_samples_F_i:
        #     raise Exception(
        #         (
        #             "Ordinary least squares problem is undetermined "
        #             "n_samples=%d must be greater than the regression"
        #             " model size p+q=%d."
        #         )
        #         % (n_samples_F_i, self.p_all[lvl] + self.q_all[lvl])
        #     )

        # Determine Gaussian Process model parameters

    def _reinterpolate(self, lvl):
        if self.options["eval_noise"] and self.options["optim_var"]:
            X = self.X
            for lvl in range(self.nlvl - 1):
                self.set_training_values(
                    X[lvl], self._predict_intermediate_values(X[lvl], lvl + 1), name=lvl
                )
            self.set_training_values(
                X[-1], self._predict_intermediate_values(X[-1], self.nlvl)
            )
            self.options["eval_noise"] = False
            self._new_train()
            self.options["eval_noise"] = True

    def _componentwise_distance(self, dx, opt=0):
        d = componentwise_distance(
            dx, self.options["corr"], self.nx, power=self.options["pow_exp_power"]
        )
        return d

    def _predict_intermediate_values(self, X, lvl, descale=True):
        """
        Evaluates the model at a set of points.
        Used for training the model at level lvl.
        Allows to relax the order problem.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        lvl : level at which the prediction is made

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        n_eval, _ = X.shape
        #        if n_features_X != self.n_features:
        #            raise ValueError("Design must be an array of n_features columns.")

        # Calculate kriging mean and variance at level 0
        mu = np.zeros((n_eval, lvl))

        # if not (self.is_continuous):
        #     X_usc = X
        if descale and self.is_continuous:
            X = (X - self.X_offset) / self.X_scale

        f = self._regression_types[self.options["poly"]](X)
        f0 = self._regression_types[self.options["poly"]](X)

        # beta = self.optimal_par[0]["beta"]
        # if self.is_continuous:
        #     dx = self._differences(X, Y=self.X_norma_all[0])
        #     d = self._componentwise_distance(dx)
        #     self.corr.theta = self.optimal_theta[0]
        #     r_ = self.corr(d).reshape(n_eval, self.nt_all[0])

        # gamma = self.optimal_par[0]["gamma"]

        # Scaled predictor
        # mu[:, 0] = (np.dot(f, beta) + np.dot(r_, gamma)).ravel()

        mu[:, 0] = self.sgp.predict_values(X)[:, 0]

        # Calculate recursively kriging mean and variance at level i
        for i in range(1, lvl):
            g = self._regression_types[self.options["rho_regr"]](X)

            if self.is_continuous:
                dx = self._differences(X, Y=self.X_norma_all[i])
                d = self._componentwise_distance(dx)
                self.corr.theta = self.optimal_theta[i]
                r_ = self.corr(d).reshape(n_eval, self.nt_all[i])

            f = np.vstack((g.T * mu[:, i - 1], f0.T))
            beta = self.optimal_par[i]["beta"]
            gamma = self.optimal_par[i]["gamma"]
            # scaled predictor
            mu[:, i] = (np.dot(f.T, beta) + np.dot(r_, gamma)).ravel()

        # scaled predictor
        if descale:
            mu = mu * self.y_std + self.y_mean

        return mu[:, -1].reshape((n_eval, 1))

    def _predict_values(self, X, is_acting=None):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """

        return self._predict_intermediate_values(X, self.nlvl)

    def _predict_variances(
        self, X: np.ndarray, is_acting=None, is_ri=False
    ) -> np.ndarray:
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        return self.predict_variances_all_levels(X)[0][:, -1]

    def predict_variances_all_levels(self, X, is_acting=None):
        # def predict_variances_all_levels(self, X):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        # Initialization X = atleast_2d(X)
        nlevel = self.nlvl
        sigma2_rhos = []
        n_eval, n_features_X = X.shape
        #        if n_features_X != self.n_features:
        #            raise ValueError("Design must be an array of n_features columns.")

        X = (X - self.X_offset) / self.X_scale

        # Calculate kriging mean and variance at level 0
        mu = np.zeros((n_eval, nlevel))

        f = self._regression_types[self.options["poly"]](X)
        f0 = self._regression_types[self.options["poly"]](X)

        # Get regression function and correlation
        # F = self.F_all[0]
        # C = self.optimal_par[0]["C"]

        # beta = self.optimal_par[0]["beta"]
        # Ft = solve_triangular(C, F, lower=True)

        # if self.is_continuous:
        #     dx = self._differences(X, Y=self.X_norma_all[0])
        #     d = self._componentwise_distance(dx)
        #     self.corr.theta = self.optimal_theta[0]
        #     r_ = self.corr(d).reshape(n_eval, self.nt_all[0])

        # gamma = self.optimal_par[0]["gamma"]

        # Scaled predictor
        # mu[:, 0] = (np.dot(f, beta) + np.dot(r_, gamma)).ravel()

        mu[:, 0] = self.sgp.predict_values(X)[:, 0]

        self.sigma2_rho = nlevel * [None]
        MSE = np.zeros((n_eval, nlevel))
        # r_t = solve_triangular(C, r_.T, lower=True)
        # G = self.optimal_par[0]["G"]

        # u_ = solve_triangular(G.T, f.T - np.dot(Ft.T, r_t), lower=True)
        # sigma2 = self.optimal_par[0]["sigma2"] / self.y_std**2
        # MSE[:, 0] = sigma2 * (
        #     # 1 + self.optimal_noise_all[0] - (r_t ** 2).sum(axis=0) + (u_ ** 2).sum(axis=0)
        #     1 - (r_t**2).sum(axis=0) + (u_**2).sum(axis=0)
        # )

        MSE[:, 0] = self.sgp.predict_variances(X)[:, 0]

        # Calculate recursively kriging variance at level i
        for i in range(1, nlevel):
            F = self.F_all[i]
            C = self.optimal_par[i]["C"]

            g = self._regression_types[self.options["rho_regr"]](X)

            if self.is_continuous:
                dx = self._differences(X, Y=self.X_norma_all[i])
                d = self._componentwise_distance(dx)
                self.corr.theta = self.optimal_theta[i]
                r_ = self.corr(d).reshape(n_eval, self.nt_all[i])

            f = np.vstack((g.T * mu[:, i - 1], f0.T))

            Ft = solve_triangular(C, F, lower=True)
            yt = solve_triangular(C, self.y_norma_all[i], lower=True)
            r_t = solve_triangular(C, r_.T, lower=True)
            G = self.optimal_par[i]["G"]
            beta = self.optimal_par[i]["beta"]

            # scaled predictor
            sigma2 = self.optimal_par[i]["sigma2"] / self.y_std**2
            q = self.q_all[i]
            u_ = solve_triangular(G.T, f - np.dot(Ft.T, r_t), lower=True)
            sigma2_rho = np.dot(
                g,
                sigma2 * linalg.inv(np.dot(G.T, G))[:q, :q]
                + np.dot(beta[:q], beta[:q].T),
            )
            sigma2_rho = (sigma2_rho * g).sum(axis=1)
            sigma2_rhos.append(sigma2_rho)

            if self.name in ["MFKPLS", "MFKPLSK"]:
                p = self.p_all[i]
                Q_ = (np.dot((yt - np.dot(Ft, beta)).T, yt - np.dot(Ft, beta)))[0, 0]
                MSE[:, i] = (
                    # sigma2_rho * MSE[:, i - 1]
                    +Q_
                    / (2 * (self.nt_all[i] - p - q))
                    # * (1 + self.optimal_noise_all[i] - (r_t ** 2).sum(axis=0))
                    * (1 - (r_t**2).sum(axis=0))
                    + sigma2 * (u_**2).sum(axis=0)
                )
            else:
                MSE[:, i] = sigma2 * (
                    1
                    + self.optimal_noise_all[i]
                    - (r_t**2).sum(axis=0)
                    + (u_**2).sum(axis=0)
                    # 1 - (r_t**2).sum(axis=0) + (u_**2).sum(axis=0)
                )  # + sigma2_rho * MSE[:, i - 1]
            if self.options["propagate_uncertainty"]:
                MSE[:, i] = MSE[:, i] + sigma2_rho * MSE[:, i - 1]

        # scaled predictor
        MSE *= self.y_std**2

        return MSE, sigma2_rhos

    def _predict_derivatives(self, x, kx):
        """
        Evaluates the derivatives at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        kx : int
            The 0-based index of the input variable with respect to which derivatives are desired.

        Returns
        -------
        y : np.ndarray*self.y_std/self.X_scale[kx])
            Derivative values.
        """

        lvl = self.nlvl
        # Initialization

        n_eval, n_features_x = x.shape
        x = (x - self.X_offset) / self.X_scale

        dy_dx = np.zeros((n_eval, lvl))

        if self.options["corr"] != "squar_exp":
            raise ValueError(
                "The derivative is only available for square exponential kernel"
            )
        if self.options["poly"] == "constant":
            df = np.zeros([n_eval, 1])
        elif self.options["poly"] == "linear":
            df = np.zeros((n_eval, self.nx + 1))
            df[:, 1:] = 1
        else:
            raise ValueError(
                "The derivative is only available for ordinary kriging or "
                + "universal kriging using a linear trend"
            )

        df0 = deepcopy(df)

        if self.options["rho_regr"] != "constant":
            raise ValueError(
                "The derivative is only available for regression rho constant"
            )
        # Get pairwise componentwise L1-distances to the input training set
        dx = self._differences(x, Y=self.X_norma_all[0])
        d = self._componentwise_distance(dx)
        # Compute the correlation function
        self.corr.theta = self.optimal_theta[0]
        r_ = self.corr(d).reshape(n_eval, self.nt_all[0])

        # Beta and gamma = R^-1(y-FBeta)
        beta = self.optimal_par[0]["beta"]
        gamma = self.optimal_par[0]["gamma"]

        df_dx = np.dot(df, beta)
        d_dx = x[:, kx].reshape((n_eval, 1)) - self.X_norma_all[0][:, kx].reshape(
            (1, self.nt_all[0])
        )

        theta = self._get_theta(0)
        dy_dx[:, 0] = np.ravel((df_dx - 2 * theta[kx] * np.dot(d_dx * r_, gamma)))

        # Calculate recursively derivative at level i
        for i in range(1, lvl):
            g = self._regression_types[self.options["rho_regr"]](x)
            dx = self._differences(x, Y=self.X_norma_all[i])
            d = self._componentwise_distance(dx)
            self.corr.theta = self.optimal_theta[i]
            r_ = self.corr(d).reshape(n_eval, self.nt_all[i])
            df = np.vstack((g.T * dy_dx[:, i - 1], df0.T))

            beta = self.optimal_par[i]["beta"]
            gamma = self.optimal_par[i]["gamma"]

            df_dx = np.dot(df.T, beta)
            d_dx = x[:, kx].reshape((n_eval, 1)) - self.X_norma_all[i][:, kx].reshape(
                (1, self.nt_all[i])
            )

            theta = self._get_theta(i)

            # scaled predictor
            dy_dx[:, i] = np.ravel(df_dx - 2 * theta[kx] * np.dot(d_dx * r_, gamma))

        deriv = dy_dx[:, -1] * self.y_std / self.X_scale[kx]
        return deriv.reshape((x.shape[0], self.ny))

    def _get_theta(self, i):
        return self.optimal_theta[i]

    def _check_param(self):
        """
        Overrides KrgBased implementation
        This function checks some parameters of the model.
        """

        if self.name in ["MFKPLS", "MFKPLSK"]:
            d = self.options["n_comp"]
        else:
            d = self.nx

        if self.options["corr"] == "act_exp":
            raise ValueError("act_exp correlation function must be used with MGP")

        if self.name in ["MFKPLS"]:
            if self.options["corr"] not in ["squar_exp", "abs_exp"]:
                raise ValueError(
                    "MFKPLS only works with a squared exponential or an absolute exponential kernel"
                )
        elif self.name in ["MFKPLSK"]:
            if self.options["corr"] not in ["squar_exp"]:
                raise ValueError(
                    "MFKPLSK only works with a squared exponential kernel (until we prove the contrary)"
                )
        # noise0 may be a list of noise values for various fi levels with various length
        max_noise = np.max([np.max(row) for row in self.options["noise0"]])
        if (self.options["eval_noise"] or max_noise > 1e-12) and self.options[
            "hyper_opt"
        ] == "TNC":
            self.options["hyper_opt"] = "Cobyla"
            warnings.warn(
                "TNC not available yet for noise handling. Switching to Cobyla"
            )
        n_param = d

        if self.options["categorical_kernel"] is not None:
            if self.options["categorical_kernel"] == MixIntKernelType.CONT_RELAX:
                n_comp = self.options["n_comp"] if "n_comp" in self.options else None
                mat_dim = (
                    self.options["cat_kernel_comps"]
                    if "cat_kernel_comps" in self.options
                    else None
                )

                n_param = compute_n_param(
                    self.design_space,
                    self.options["categorical_kernel"],
                    d,
                    n_comp,
                    mat_dim,
                )
            else:
                raise ValueError(
                    "Only the continuous relaxation kernel is available with multi-fidelity"
                )

        if isinstance(self.options["theta0"], np.ndarray):
            if self.options["theta0"].shape != (self.nlvl, n_param):
                raise ValueError(
                    "the dimensions of theta0 %s should coincide to the number of dim %s"
                    % (self.options["theta0"].shape, (self.nlvl, n_param))
                )
        else:
            if len(self.options["theta0"]) != n_param:
                if len(self.options["theta0"]) == 1:
                    self.options["theta0"] *= np.ones((self.nlvl, n_param))
                elif len(self.options["theta0"]) == self.nlvl:
                    self.options["theta0"] = np.array(self.options["theta0"]).reshape(
                        -1, 1
                    )
                    self.options["theta0"] *= np.ones((1, n_param))
                else:
                    raise ValueError(
                        "the length of theta0 (%s) should be equal to the number of dim (%s) \
                            or levels of fidelity (%s)."
                        % (len(self.options["theta0"]), n_param, self.nlvl)
                    )
            else:
                self.options["theta0"] *= np.ones((self.nlvl, 1))

        if len(self.options["noise0"]) != self.nlvl:
            if len(self.options["noise0"]) == 1:
                self.options["noise0"] = self.nlvl * [self.options["noise0"]]
            else:
                raise ValueError(
                    "the length of noise0 (%s) should be equal to the number of levels of fidelity (%s)."
                    % (len(self.options["noise0"]), self.nlvl)
                )

        for i in range(self.nlvl):
            if self.options["use_het_noise"]:
                if len(self.X[i]) == len(np.unique(self.X[i])):
                    if len(self.options["noise0"][i]) != self.nt_all[i]:
                        if len(self.options["noise0"][i]) == 1:
                            self.options["noise0"][i] *= np.ones(self.nt_all[i])
                        else:
                            raise ValueError(
                                "for the level of fidelity %s, the length of noise0 (%s) should be equal to \
                                    the number of observations (%s)."
                                % (i, len(self.options["noise0"][i]), self.nt_all[i])
                            )
            else:
                if np.size(self.options["noise0"][i]) != 1:
                    raise ValueError(
                        "for the level of fidelity %s, the length of noise0 (%s) should be equal to one."
                        % (i, len(self.options["noise0"][i]))
                    )



# -*- coding: utf-8 -*-
"""
Created on Sat May 04 10:10:12 2024

@author: Mauricio Castano Aguirre <mauricio.castano_aguirre@onera.fr>
Multi-Fidelity co-Kriging model construction for non-nested experimental
design sets.
-------
[1] Loic Le Gratiet (2013). Multi-fidelity Gaussian process modelling
[Doctoral Thesis, Université Paris-Sud].
[2] Edwin V. Bonilla, Kian Ming A. Chai, and Christopher K. I. Williams
(2007). Multi-task Gaussian Process prediction. In International
Conference on Neural Information Processing Systems.
"""

# import warnings
import numpy as np
from scipy.linalg import solve_triangular
from scipy import optimize
from smt.sampling_methods import LHS
from smt.utils.kriging import differences, componentwise_distance
from smt.surrogate_models.krg_based import KrgBased
from smt.utils.misc import standardization
from scipy.cluster.vq import kmeans


class SMFCK(KrgBased):
    def _initialize(self):
        super()._initialize()
        declare = self.options.declare
        self.name = "SMFCK"

        declare(
            "rho0",
            1.0,
            types=(float),
            desc="Initial rho for the autoregressive model , \
                  (scalar factor between two consecutive fidelities, \
                    e.g., Y_HF = (Rho) * Y_LF + Gamma",
        )
        declare(
            "rho_bounds",
            [-5.0, 5.0],
            types=(list, np.ndarray),
            desc="Bounds for the rho parameter used in the autoregressive model",
        )
        declare(
            "sigma0",
            1.0,
            types=(float),
            desc="Initial variance parameter",
        )
        declare(
            "sigma_bounds",
            [1e-6, 100],
            types=(list, np.ndarray),
            desc="Bounds for the variance parameter",
        )
        declare(
            "lambda",
            0.0,
            types=(float),
            desc="Regularization parameter",
        )
        declare(
            "hyper_opt",
            "Cobyla",
            values=("Cobyla", "Cobyla-nlopt"),
            desc="Optimiser for hyperparameters optimisation",
        )
        declare(
            "n_inducing",
            [6, 5],
            types=(list, np.ndarray),
            desc="Number of inducing points per fidelity level",
        )
        declare(
            "method",
            "FITC",
            values=("FITC"),
            desc="Methods available for Sparse Multi-fidelity",
            types=(str),
        )
        declare(
            "nugget",
            1000.0
            * np.finfo(
                np.double
            ).eps,  # slightly increased compared to kriging-based one
            types=(float),
            desc="a jitter for numerical stability",
        )
        declare(
            "inducing_method",
            "kmeans",
            types=str,
            values=["random", "kmeans"],
            desc="The chosen method to induce points",
        )
        self.options["hyper_opt"] = (
            "Cobyla-nlopt"  # MFCK doesn't support gradient-based optimizers
        )

    def train(self):
        """
        Overrides MFK implementation
        Trains the Multi-Fidelity co-Kriging model
        Returns
        -------
        None.
        """
        xt = []
        yt = []
        zt = []
        i = 0
        while self.training_points.get(i, None) is not None:
            xt.append(self.training_points[i][0][0])
            yt.append(self.training_points[i][0][1])

            if self.options["inducing_method"] == "random":
                idx = np.random.permutation(self.nt)[: self.options["n_inducing"][i]]
                zt.append(xt[idx])
            elif self.options["inducing_method"] == "kmeans":
                zt.append(
                    kmeans(
                        self.training_points[i][0][0], self.options["n_inducing"][i]
                    )[0]
                )
            i = i + 1
        xt.append(self.training_points[None][0][0])
        yt.append(self.training_points[None][0][1])

        if self.options["inducing_method"] == "random":
            idx = np.random.permutation(self.nt)[: self.options["n_inducing"][i]]
            zt.append(xt[idx])
        elif self.options["inducing_method"] == "kmeans":
            zt.append(
                kmeans(self.training_points[None][0][0], self.options["n_inducing"][i])[
                    0
                ]
            )
        # zt.append(kmeans(self.training_points[None][0][0],self.options["n_inducing"][i])[0])
        self.lvl = i + 1
        self.X = xt
        self.Z = zt

        if np.shape(self.options["n_inducing"])[0] == self.lvl:
            self.y = np.vstack(yt)
            self._check_param()

            (
                _,
                _,
                self.X_offset,
                self.y_mean,
                self.X_scale,
                self.y_std,
            ) = standardization(np.concatenate(xt, axis=0), np.concatenate(yt, axis=0))

            self.X_norma_all = [(x - self.X_offset) / self.X_scale for x in xt]
            self.Z_norma_all = [(x - self.X_offset) / self.X_scale for x in zt]
            self.y_norma_all = np.vstack([(f - self.y_mean) / self.y_std for f in yt])

        else:
            raise ValueError(
                f"n_inducing {self.options['n_inducing']} don't correspond to the fidelities"
            )

        if self.lvl == 1:
            # For a single level, initialize theta_ini, lower_bounds, and
            # upper_bounds with consistent shapes
            theta_ini = np.hstack(
                (self.options["sigma0"], self.options["theta0"])
            )  # Variance + initial theta values
            lower_bounds = np.hstack(
                (
                    self.options["sigma_bounds"][0],
                    np.full(self.nx, self.options["theta_bounds"][0]),
                )
            )
            upper_bounds = np.hstack(
                (
                    self.options["sigma_bounds"][1],
                    np.full(self.nx, self.options["theta_bounds"][1]),
                )
            )
            # Apply log10 to theta_ini and bounds
            nb_params = len(self.options["theta0"])
            theta_ini[: nb_params + 1] = np.log10(theta_ini[: nb_params + 1])
            lower_bounds[: nb_params + 1] = np.log10(lower_bounds[: nb_params + 1])
            upper_bounds[: nb_params + 1] = np.log10(upper_bounds[: nb_params + 1])
        else:
            for lvl in range(self.lvl):
                if lvl == 0:
                    # Initialize theta_ini for level 0
                    theta_ini = np.hstack(
                        (self.options["sigma0"], self.options["theta0"])
                    )  # Variance + initial theta values
                    lower_bounds = np.hstack(
                        (
                            self.options["sigma_bounds"][0],
                            np.full(self.nx, self.options["theta_bounds"][0]),
                        )
                    )
                    upper_bounds = np.hstack(
                        (
                            self.options["sigma_bounds"][1],
                            np.full(self.nx, self.options["theta_bounds"][1]),
                        )
                    )
                    # Apply log10 to theta_ini and bounds
                    nb_params = len(self.options["theta0"])
                    theta_ini[: nb_params + 1] = np.log10(theta_ini[: nb_params + 1])
                    lower_bounds[: nb_params + 1] = np.log10(
                        lower_bounds[: nb_params + 1]
                    )
                    upper_bounds[: nb_params + 1] = np.log10(
                        upper_bounds[: nb_params + 1]
                    )

                elif lvl > 0:
                    # For additional levels, append to theta_ini, lower_bounds, and upper_bounds
                    thetat = np.hstack((self.options["sigma0"], self.options["theta0"]))
                    lower_boundst = np.hstack(
                        (
                            self.options["sigma_bounds"][0],
                            np.full(self.nx, self.options["theta_bounds"][0]),
                        )
                    )
                    upper_boundst = np.hstack(
                        (
                            self.options["sigma_bounds"][1],
                            np.full(self.nx, self.options["theta_bounds"][1]),
                        )
                    )
                    # Apply log10 to the newly added values
                    thetat = np.log10(thetat)
                    lower_boundst = np.log10(lower_boundst)
                    upper_boundst = np.log10(upper_boundst)
                    # Append to theta_ini, lower_bounds, and upper_bounds
                    theta_ini = np.hstack([theta_ini, thetat, self.options["rho0"]])
                    lower_bounds = np.hstack([lower_bounds, lower_boundst])
                    upper_bounds = np.hstack([upper_bounds, upper_boundst])
                    # Finally, append the rho bounds
                    lower_bounds = np.hstack(
                        [lower_bounds, self.options["rho_bounds"][0]]
                    )
                    upper_bounds = np.hstack(
                        [upper_bounds, self.options["rho_bounds"][1]]
                    )

        if self.options["eval_noise"]:
            theta_ini = np.hstack(
                [theta_ini, np.full(self.lvl, self.options["noise0"][0])]
            )

            lower_bounds = np.hstack(
                [lower_bounds, np.full(self.lvl, self.options["noise_bounds"][0])]
            )

            upper_bounds = np.hstack(
                [upper_bounds, np.full(self.lvl, self.options["noise_bounds"][1])]
            )

        theta_ini = theta_ini[:].T

        theta_ini[-self.lvl : :] = np.log10(theta_ini[-self.lvl : :])
        upper_bounds[-self.lvl : :] = np.log10(upper_bounds[-self.lvl : :])
        lower_bounds[-self.lvl : :] = np.log10(lower_bounds[-self.lvl : :])

        x_opt = theta_ini

        if self.options["hyper_opt"] == "Cobyla":
            if self.options["n_start"] > 1:
                sampling = LHS(
                    xlimits=np.stack((lower_bounds, upper_bounds), axis=1),
                    criterion="ese",
                    random_state=0,
                )
                theta_lhs_loops = sampling(self.options["n_start"])
                theta0 = np.vstack((theta_ini, theta_lhs_loops))
            else:
                theta0 = np.vstack((theta_ini, theta_ini))
            constraints = []

            for i in range(len(theta_ini)):
                constraints.append(lambda theta0, i=i: theta0[i] - lower_bounds[i])
                constraints.append(lambda theta0, i=i: upper_bounds[i] - theta0[i])

            for j in range(self.options["n_start"]):
                optimal_theta_res_loop = optimize.minimize(
                    self.neg_log_likelihood_scipy,
                    theta0[j, :],
                    method="COBYLA",
                    constraints=[{"fun": con, "type": "ineq"} for con in constraints],
                    options={
                        "rhobeg": 0.5,
                        "tol": 1e-6,
                        "maxiter": 50,
                    },
                )
                x_opt_iter = optimal_theta_res_loop.x

                if j == 0:
                    x_opt = x_opt_iter
                    nll = optimal_theta_res_loop["fun"]
                else:
                    if optimal_theta_res_loop["fun"] < nll:
                        x_opt = x_opt_iter
                        nll = optimal_theta_res_loop["fun"]

        elif self.options["hyper_opt"] == "Cobyla-nlopt":
            try:
                import nlopt
            except ImportError:
                print("nlopt library is not installed or available on this system")

            opt = nlopt.opt(nlopt.LN_COBYLA, theta_ini.shape[0])
            opt.set_lower_bounds(lower_bounds)  # Lower bounds for each dimension
            opt.set_upper_bounds(upper_bounds)  # Upper bounds for each dimension
            opt.set_min_objective(self.neg_log_likelihood_nlopt)
            opt.set_maxeval(80)
            opt.set_xtol_rel(1e-6)
            x0 = np.copy(theta_ini)
            x_opt = opt.optimize(x0)
        else:
            raise ValueError(
                f"The optimizer {self.options['hyper_opt']} is not available"
            )

        if self.options["eval_noise"]:
            x_opt1 = np.array(x_opt[: -self.lvl], copy=True)
        else:
            x_opt1 = np.array(x_opt, copy=True)

        x_opt1[0] = 10 ** (x_opt1[0])  # Apply 10** to Sigma 0
        x_opt1[1 : self.nx + 1] = (
            10 ** (x_opt1[1 : self.nx + 1])
        )  # Apply 10** to length scales 0

        x_opt1[self.nx + 1 :: self.nx + 2] = (
            10 ** (x_opt1[self.nx + 1 :: self.nx + 2])
        )  # Apply 10** to sigmas gamma

        for i in np.arange(self.nx + 2, x_opt1.shape[0] - 1, self.nx + 2):
            x_opt1[i : i + self.nx] = 10 ** x_opt1[i : i + self.nx]

        if self.options["eval_noise"]:
            x_opt = np.array(x_opt, copy=True)

            x_opt[-self.lvl : :] = 10 ** x_opt[-self.lvl : :]

            x_opt[: -self.lvl] = x_opt1
        else:
            x_opt = np.array(x_opt, copy=True)
            x_opt = x_opt1
        self.optimal_theta = x_opt

    def eta(self, j, jp, rho):
        """Compute eta_{j,l} based on the given rho values."""
        if j < jp:
            return np.prod(rho[j:jp])  # Product of rho[j+1] to rho[l]
        elif j == jp:
            return 1
        else:
            raise ValueError(
                f"The iterative variable j={j} cannot be greater than j'={jp}"
            )

    # Covariance between y_l(x) and y_l'(x')
    def compute_cross_K(self, x, xp, L, Lp, param):
        """
        Calculation Cov(y_l(x), y_{l'}(x')) using the autoregressive formulation.
        Parmeters:
        - x: First input for the covariannce (np.ndarray)
        - xp: Second input for the covariannce (np.ndarray)
        - L: Level index of the first output (scalar)
        - Lp: Level index of the second output (scalar)
        - param: Set of Hyper-parameters (vector)
        Returns:
        - Covariance matrix cov(y_l(x), y_{l'}(x')) (np.ndarray)
        """
        cov_value = 0.0

        sigma_0 = param[0]
        l_0 = param[1 : self.nx + 1]
        # param0 = param[0 : self.nx+1]
        sigmas_gamma = param[self.nx + 1 :: self.nx + 2]
        l_s = [
            param[i : i + self.nx].tolist()
            for i in np.arange(self.nx + 2, param.shape[0] - 1, self.nx + 2)
        ]
        # ls_gamma = param[3::3]
        rho_values = param[2 + 2 * self.nx :: self.nx + 2]

        # Sum of j=0 until l_^prime
        for j in range(Lp + 1):
            eta_j_l = self.eta(j, L, rho_values)
            eta_j_lp = self.eta(j, Lp, rho_values)

            if j == 0:
                # Cov(γ_j(x), γ_j(x')) using the kernel for K_00
                cov_gamma_j = self._compute_K(x, xp, [sigma_0, l_0])
            else:
                # Cov(γ_j(x), γ_j(x')) using the kernel
                cov_gamma_j = self._compute_K(x, xp, [sigmas_gamma[j - 1], l_s[j - 1]])
            # Add to the value of the covariance
            cov_value += eta_j_l * eta_j_lp * cov_gamma_j

        return cov_value

    def predict_all_levels(self, x):
        """
        Generalized prediction function for the multi-fidelity co-Kriging
        Parameters
        ----------
        x : np.ndarray
            Array with the inputs for make the prediction.
        Returns
        -------
        means : (list, np.array)
            Returns the conditional means per level.
        covariances: (list, np.array)
            Returns the conditional covariance matrixes per level.
        """
        means = []
        covariances = []
        x = (x - self.X_offset) / self.X_scale

        if self.lvl == 1:
            sigma = self.optimal_theta[0]
            l_s = [self.optimal_theta[1 : self.nx + 1]]

            k_XX = self._compute_K(
                self.X_norma_all[0], self.X_norma_all[0], [sigma, l_s[0]]
            )
            k_xX = self._compute_K(x, self.X_norma_all[0], [sigma, l_s[0]])
            k_xx = self._compute_K(x, x, [sigma, l_s[0]])

            L = np.linalg.cholesky(
                k_XX + self.options["nugget"] * np.eye(k_XX.shape[0])
            )

            beta1 = solve_triangular(L, k_xX.T, lower=True)
            alpha1 = solve_triangular(L, self.y_norma_all, lower=True)
            means.append(self.y_std * np.dot(beta1.T, alpha1) + self.y_mean)
            covariances.append(k_xx - np.dot(beta1.T, beta1))
        else:
            self.K = self.compute_blockwise_K(
                self.X_norma_all, self.X_norma_all, self.optimal_theta[: -self.lvl]
            )

            noises = self.optimal_theta[-self.lvl : :]
            k_xX = []
            k_xZ = []
            for ind in range(self.lvl):
                k_xx = self.compute_cross_K(
                    x, x, ind, ind, self.optimal_theta[: -self.lvl]
                )

                noise_matrix = noises[ind] * np.eye(k_xx.shape[0])

                for j in range(self.lvl):
                    if ind >= j:
                        k_xZ.append(
                            self.compute_cross_K(
                                self.Z_norma_all[j],
                                x,
                                ind,
                                j,
                                self.optimal_theta[: -self.lvl],
                            )
                        )
                    else:
                        k_xZ.append(
                            self.compute_cross_K(
                                self.Z_norma_all[j],
                                x,
                                j,
                                ind,
                                self.optimal_theta[: -self.lvl],
                            )
                        )
                means.append(
                    self.y_std * (np.vstack(k_xZ).T @ self.woodbury_vec) + self.y_mean
                )
                val = np.sum(
                    np.dot(self.woodbury_inv.T, np.vstack(k_xZ)) * np.vstack(k_xZ), 0
                )
                var = (np.diag(k_xx + noise_matrix) - val)[:, None]
                var = np.clip(var, 1e-15, np.inf)
                # var += (noises[ind]*np.eye(k_xx.shape[0]))
                covariances.append(var * self.y_std**2)
                k_xZ.clear()
        # print("Optimal noise",noises * self.y_std**2)
        return means, covariances

    def predict_values(self, x, is_acting=None):
        """
        Prediction function for the highest fidelity level
        Parameters
        ----------
        x : array
            Array with the inputs for make the prediction.
        Returns
        -------
        mean : np.array
            Returns the conditional means per level.
        covariance: np.ndarray
            Returns the conditional covariance matrixes per level.
        """
        means, covariances = self.predict_all_levels(x)

        return means[self.lvl - 1]

    def predict_variances(
        self, X: np.ndarray, is_acting=None, is_ri=False
    ) -> np.ndarray:
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        means, covariances = self.predict_all_levels(X)

        return covariances[self.lvl - 1]

    def neg_log_likelihood(self, param, grad=None):
        if self.lvl == 1:
            sigma = param[0]
            l_s = [param[1 : self.nx + 1]]
            self.K = self._compute_K(self.X[0], self.X[0], [sigma, l_s])
        else:
            self.K = self.compute_blockwise_K(
                self.X_norma_all, self.X_norma_all, param[: -self.lvl]
            )
        if self.options["eval_noise"]:
            noises = param[-self.lvl : :]
            varis = []
            for i, v in enumerate(noises):
                varis = np.hstack([varis, np.full(self.X[i].shape[0], noises[i])])
            noise_matrix = varis * np.eye(self.K.shape[0])
        else:
            noise_matrix = self.options["nugget"] * np.eye(self.K.shape[0])

        if self.options["method"] == "FITC":
            Kmm = self.compute_blockwise_K(
                self.Z_norma_all, self.Z_norma_all, param[: -self.lvl]
            )
            Knm = self.compute_blockwise_K(
                self.X_norma_all, self.Z_norma_all, param[: -self.lvl]
            )

            U = np.linalg.cholesky(Kmm + np.eye(Kmm.shape[0]) * self.options["nugget"])

            Ui = np.linalg.inv(U)
            V = Ui @ Knm.T

            nu = np.diag(self.K) - np.sum(np.square(V), 0) + np.diag(noise_matrix)

            beta = 1.0 / nu

            A = np.eye(Kmm.shape[0]) + V * beta @ V.T
            L = np.linalg.cholesky(A)
            Li = np.linalg.inv(L)
            a = np.einsum("ij,i->ij", self.y_norma_all, beta)
            b = Li @ V @ a

            likelihood = 0.5 * (
                +np.sum(np.log(nu))
                + 2.0 * np.sum(np.log(np.diag(L)))
                + a.T @ self.y_norma_all
                - np.einsum("ij,ij->", b, b)
            )

            LiUi = Li @ Ui
            LiUiT = LiUi.T
            self.woodbury_vec = LiUiT @ b
            self.woodbury_inv = Ui.T @ Ui - LiUiT @ LiUi
        # Temporary unavailable
        # elif self.options["method"]=="VFE":
        #     Kmm = self.compute_blockwise_K(self.Z_norma_all, self.Z_norma_all, param[:-self.lvl])
        #     Knm = self.compute_blockwise_K(self.X_norma_all, self.Z_norma_all, param[:-self.lvl])
        #     U = np.linalg.cholesky(Kmm + np.eye(Kmm.shape[0]) * self.options["nugget"])
        #     Ui = np.linalg.inv(U)
        #     V = Ui @ Knm.T
        #     nu = np.diag(self.K) - np.sum(np.square(V), 0) + np.diag(noise_matrix)
        #     val = np.trace(self.K + Knm @ V  )
        #     beta = 1.0 / nu
        #     A = np.eye(Kmm.shape[0]) + V * beta @ V.T
        #     L = np.linalg.cholesky(A)
        #     Li = np.linalg.inv(L)
        #     a = np.einsum("ij,i->ij", self.y_norma_all, beta)
        #     b = Li @ V @ a
        #     likelihood = 0.5 * (
        #         +val
        #         +np.sum(np.log(nu))
        #         + 2.0 * np.sum(np.log(np.diag(L)))
        #         + a.T @ self.y_norma_all
        #         - np.einsum("ij,ij->", b, b)
        #     )
        #     LiUi = Li @ Ui
        #     LiUiT = LiUi.T
        #     self.woodbury_vec = LiUiT @ b
        #     self.woodbury_inv = Ui.T @ Ui - LiUiT @ LiUi
        return likelihood[0][0]

    def neg_log_likelihood_scipy(self, param):
        """
        Likelihood for Cobyla-scipy (SMT) optimizer
        """
        if self.options["eval_noise"]:
            param1 = np.array(param[: -self.lvl], copy=True)
        else:
            param1 = np.array(param, copy=True)

        param1[0] = 10 ** (param[0])  # Apply 10** to Sigma 0
        param1[1 : self.nx + 1] = (
            10 ** (param[1 : self.nx + 1])
        )  # Apply 10** to length scales 0
        param1[self.nx + 1 :: self.nx + 2] = (
            10 ** (param1[self.nx + 1 :: self.nx + 2])
        )  # Apply 10** to sigmas gamma

        for i in np.arange(self.nx + 2, param1.shape[0] - 1, self.nx + 2):
            param1[i : i + self.nx] = 10 ** param1[i : i + self.nx]

        if self.options["eval_noise"]:
            param = np.array(param, copy=True)

            param[-self.lvl : :] = 10 ** param[-self.lvl : :]

            param[: -self.lvl] = param1
        else:
            param = np.array(param, copy=True)
            param = param1
        return self.neg_log_likelihood(param)

    def neg_log_likelihood_nlopt(self, param, grad=None):
        """
        Likelihood for nlopt optimizers
        """
        if self.options["eval_noise"]:
            param1 = np.array(param[: -self.lvl], copy=True)
        else:
            param1 = np.array(param, copy=True)

        param1[0] = 10 ** (param[0])  # Apply 10** to Sigma 0
        param1[1 : self.nx + 1] = (
            10 ** (param[1 : self.nx + 1])
        )  # Apply 10** to length scales 0
        param1[self.nx + 1 :: self.nx + 2] = (
            10 ** (param1[self.nx + 1 :: self.nx + 2])
        )  # Apply 10** to sigmas gamma

        for i in np.arange(self.nx + 2, param1.shape[0] - 1, self.nx + 2):
            param1[i : i + self.nx] = 10 ** param1[i : i + self.nx]

        if self.options["eval_noise"]:
            param = np.array(param, copy=True)
            param[-self.lvl : :] = 10 ** param[-self.lvl : :]
            param[: -self.lvl] = param1
        else:
            param = np.array(param, copy=True)
            param = param1
        return self.neg_log_likelihood(param, grad)

    def compute_blockwise_K(self, X, Xprime, param):
        K_block = {}
        n = 0
        nprime = 0
        for i in X:
            n = n + i.shape[0]

        for i in Xprime:
            nprime = nprime + i.shape[0]
        for jp in range(self.lvl):
            for j in range(self.lvl):
                if jp >= j:
                    K_block[str(jp) + str(j)] = self.compute_cross_K(
                        X[j], Xprime[jp], jp, j, param
                    )
                else:
                    K_block[str(jp) + str(j)] = self.compute_cross_K(
                        X[j], Xprime[jp], j, jp, param
                    )
        K = np.zeros((n, nprime))
        row_init, col_init = 0, 0
        for j in range(self.lvl):
            col_init = 0
            for jp in range(self.lvl):
                r, c = K_block[str(jp) + str(j)].shape
                K[row_init : row_init + r, col_init : col_init + c] = K_block[
                    str(jp) + str(j)
                ]
                col_init += c
            row_init += r

        return K

    def _compute_K(self, A: np.ndarray, B: np.ndarray, param):
        """
        Compute the covariance matrix K between A and B
            Modified for MFCK initial test (Same theta for each dimmension)
        """
        # Compute pairwise componentwise L1-distances between A and B
        dx = differences(A, B)
        d = componentwise_distance(
            dx,
            self.options["corr"],
            self.X[0].shape[1],
            power=self.options["pow_exp_power"],
        )
        self.corr.theta = np.asarray(param[1])
        r = self.corr(d)
        R = r.reshape(A.shape[0], B.shape[0])
        K = param[0] * R
        return K



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 21 2024

@author: Lisa Pretsch <<lisa.pretsch@tum.de>>

Based on KRG from SMT by Dr. Mohamed Amine Bouhlel

With cooperative components Kriging model fit inspired by
distributed multi-disciplinary design optimization (MDO) approaches and
Zhan, D., Wu, J., Xing, H. et al.
A cooperative approach to efficient global optimization.
J Glob Optim 88, 327–357 (2024). https://doi.org/10.1007/s10898-023-01316-6
"""

from copy import deepcopy

import numpy as np
from scipy import optimize
from scipy.stats import multivariate_normal as m_norm

from smt.sampling_methods import LHS
from smt.surrogate_models.krg_based import KrgBased
from smt.utils.kriging import componentwise_distance


class CoopCompKRG(KrgBased):
    name = "Cooperative Components Kriging"

    """
    # Example with random components
    # (use physical components if available)
    n_comp = 3

    # Random design variable to component allocation
    comps = [*range(n_comp)]
    vars = [*range(n_dim)]
    random.shuffle(vars)
    comp_var = np.full((n_dim, n_comp), False)
    for c in comps:
        comp_size = int(n_dim/n_comp)
        start = c*comp_size
        end = (c+1)*comp_size
        if c+1 == n_comp:
            end = max((c+1)*comp_size, n_dim)
        comp_var[vars[start:end],c] = True

    # Cooperative components Kriging model fit
    model = CoopCompKRG(hyper_opt='Cobyla')
    for active_coop_comp in comps:
        model.set_training_values(x_train, y_train)
        model.train(active_coop_comp, comp_var)

    # Prediction as for ordinary Kriging
    y_pred = model.predict_values(x_test)
    """

    def _initialize(self):
        super()._initialize()
        declare = self.options.declare
        declare(
            "corr",
            "squar_exp",
            values=(
                "pow_exp",
                "abs_exp",
                "squar_exp",
                "squar_sin_exp",
                "matern52",
                "matern32",
            ),
            desc="Correlation function type",
            types=(str),
        )
        declare(
            "hyper_opt",
            "Cobyla",
            values=("Cobyla",),
            desc="Correlation function type",
            types=(str),
        )
        supports = self.supports
        supports["variances"] = True
        supports["derivatives"] = False
        supports["variance_derivatives"] = False
        supports["x_hierarchy"] = False

    def _componentwise_distance(self, dx, opt=0, theta=None, return_derivative=False):
        d = componentwise_distance(
            dx,
            self.options["corr"],
            self.nx,
            self.options["pow_exp_power"],
            theta=theta,
            return_derivative=return_derivative,
        )
        return d

    def _compute_pls(self, X, y):
        """
        Workaround for function call in _new_train()
        """
        return X, y

    def train(self, active_coop_comp, comp_var) -> None:
        """
        Train the model
        Overrides SurrogateModel implementation with additional parameters
        """

        # Set allocations of design variables to components
        if not isinstance(comp_var, np.ndarray) or comp_var.shape[0] != self.nx:
            raise ValueError("comp_var has the wrong data type or shape.")
        self.comp_var = [var for var in comp_var.T]

        # Set context vector to current best hyperparameter (used for inactive components)
        try:
            self.coop_theta = self.optimal_theta
        except AttributeError:
            self.coop_theta = self.options["theta0"] * np.ones((self.nx))

        # Hyperparameter optimization for active components
        self.active_comp_var = self.comp_var[active_coop_comp]
        self.num_active = np.count_nonzero(self.active_comp_var)
        super().train()

    def _optimize_hyperparam(self, D):
        """
        This function evaluates the Gaussian Process model at x.
        Overrides KrgBased implementation for component-wise cooperative
        hyperparameter optimization.

        Parameters
        ----------
        D: np.ndarray [n_obs * (n_obs - 1) / 2, dim]
            - The componentwise cross-spatial-correlation-distance between the
              vectors in X.

        Returns
        -------
        best_optimal_rlf_value: real
            - The value of the reduced likelihood function associated to the
              best autocorrelation parameters theta.
        best_optimal_par: dict()
            - A dictionary containing the requested Gaussian Process model
              parameters.
        best_optimal_theta: list(n_comp) or list(dim)
            - The best hyperparameters found by the optimization.
        """
        # reinitialize optimization best values
        self.best_iteration_fail = None
        self._thetaMemory = None
        # Initialize the hyperparameter-optimization
        if self.name in ["MGP"]:

            def minus_cooperative_reduced_likelihood_function(active_theta):
                theta = self.coop_theta
                theta[self.active_comp_var] = active_theta
                res = -self._reduced_likelihood_function(theta)[0]
                return res

            def grad_minus_cooperative_reduced_likelihood_function(active_theta):
                theta = self.coop_theta
                theta[self.active_comp_var] = active_theta
                grad = -self._reduced_likelihood_gradient(theta)[0]
                return grad

        else:

            def minus_cooperative_reduced_likelihood_function(active_log10t):
                theta = self.coop_theta
                theta[self.active_comp_var] = 10.0**active_log10t
                return -self._reduced_likelihood_function(theta)[0]

            def grad_minus_cooperative_reduced_likelihood_function(active_log10t):
                theta = self.coop_theta
                theta[self.active_comp_var] = 10.0**active_log10t
                theta_2d = np.atleast_2d(theta).T
                res = (
                    -np.log(10.0)
                    * (theta_2d)
                    * (self._reduced_likelihood_gradient(theta_2d)[0])
                )
                return res

        limit, _rhobeg = 10 * self.num_active, 0.5  # consider only active dimensions
        exit_function = False
        if "KPLSK" in self.name:
            n_iter = 1
        else:
            n_iter = 0

        (
            best_optimal_theta,
            best_optimal_rlf_value,
            best_optimal_par,
            constraints,
        ) = (
            [],
            [],
            [],
            [],
        )

        for ii in range(n_iter, -1, -1):
            bounds_hyp = []

            self.theta0 = deepcopy(self.options["theta0"])
            for i in range(
                self.num_active
            ):  # range(len(self.theta0)): # set only active constraints
                # In practice, in 1D and for X in [0,1], theta^{-2} in [1e-2,infty),
                # i.e. theta in (0,1e1], is a good choice to avoid overfitting.
                # By standardising X in R, X_norm = (X-X_mean)/X_std, then
                # X_norm in [-1,1] if considering one std intervals. This leads
                # to theta in (0,2e1]
                theta_bounds = self.options["theta_bounds"]
                if self.theta0[i] < theta_bounds[0] or self.theta0[i] > theta_bounds[1]:
                    self.theta0[i] = np.random.rand()
                    self.theta0[i] = (
                        self.theta0[i] * (theta_bounds[1] - theta_bounds[0])
                        + theta_bounds[0]
                    )
                    print(
                        "Warning: theta0 is out the feasible bounds. A random initialisation is used instead."
                    )

                if self.name in ["MGP"]:  # to be discussed with R. Priem
                    constraints.append(lambda theta, i=i: theta[i] + theta_bounds[1])
                    constraints.append(lambda theta, i=i: theta_bounds[1] - theta[i])
                    bounds_hyp.append((-theta_bounds[1], theta_bounds[1]))
                else:
                    log10t_bounds = np.log10(theta_bounds)
                    constraints.append(lambda log10t, i=i: log10t[i] - log10t_bounds[0])
                    constraints.append(lambda log10t, i=i: log10t_bounds[1] - log10t[i])
                    bounds_hyp.append(log10t_bounds)

            if self.name in ["MGP"]:
                theta0_rand = m_norm.rvs(
                    self.options["prior"]["mean"] * len(self.theta0),
                    self.options["prior"]["var"],
                    1,
                )
                theta0 = self.theta0
            else:
                theta0_rand = np.random.rand(len(self.theta0))
                theta0_rand = (
                    theta0_rand * (log10t_bounds[1] - log10t_bounds[0])
                    + log10t_bounds[0]
                )
                theta0 = np.log10(self.theta0)
            ##from abs distance to kernel distance
            self.D = self._componentwise_distance(D, opt=ii)

            # Initialization
            k, stop, best_optimal_rlf_value = 0, 1, -1e20
            while k < stop:
                # Use specified starting point as first guess
                self.noise0 = np.array(self.options["noise0"])
                noise_bounds = self.options["noise_bounds"]
                if self.options["eval_noise"] and not self.options["use_het_noise"]:
                    self.noise0[self.noise0 == 0.0] = noise_bounds[0]
                    for i in range(len(self.noise0)):
                        if (
                            self.noise0[i] < noise_bounds[0]
                            or self.noise0[i] > noise_bounds[1]
                        ):
                            self.noise0[i] = noise_bounds[0]
                            print(
                                "Warning: noise0 is out the feasible bounds. The lowest possible value is used instead."
                            )

                    theta0 = np.concatenate(
                        [theta0, np.log10(np.array([self.noise0]).flatten())]
                    )
                    theta0_rand = np.concatenate(
                        [
                            theta0_rand,
                            np.log10(np.array([self.noise0]).flatten()),
                        ]
                    )

                    for i in range(len(self.noise0)):
                        noise_bounds = np.log10(noise_bounds)
                        constraints.append(
                            lambda log10t: log10t[i + len(self.theta0)]
                            - noise_bounds[0]
                        )
                        constraints.append(
                            lambda log10t: noise_bounds[1]
                            - log10t[i + len(self.theta0)]
                        )
                        bounds_hyp.append(noise_bounds)
                theta_limits = np.repeat(
                    np.log10([theta_bounds]), repeats=len(theta0), axis=0
                )
                theta_all_loops = np.vstack((theta0, theta0_rand))

                if self.options["n_start"] > 1:
                    sampling = LHS(
                        xlimits=theta_limits, criterion="maximin", random_state=41
                    )
                    theta_lhs_loops = sampling(self.options["n_start"])
                    theta_all_loops = np.vstack((theta_all_loops, theta_lhs_loops))

                optimal_theta_res = {"fun": float("inf")}
                try:
                    if (
                        self.options["hyper_opt"] == "Cobyla"
                        and self.options["n_start"] >= 1
                    ):
                        for theta0_loop in theta_all_loops:
                            optimal_theta_res_loop = optimize.minimize(
                                minus_cooperative_reduced_likelihood_function,  # adjusted likelihood function
                                theta0_loop[
                                    self.active_comp_var
                                ],  # only optimize part of theta
                                constraints=[  # only consider part of the constraints
                                    {"fun": con, "type": "ineq"} for con in constraints
                                ],
                                method="COBYLA",
                                options={
                                    "rhobeg": _rhobeg,
                                    "tol": 1e-4,
                                    "maxiter": limit,
                                },
                            )
                            if self.name not in ["MGP"]:
                                optimal_coop_theta = np.log10(self.coop_theta)
                            else:
                                optimal_coop_theta = self.coop_theta
                            optimal_coop_theta[self.active_comp_var] = (
                                optimal_theta_res_loop["x"]
                            )
                            # augment "x" in optimal_theta_res_loop with theta_best of other indices
                            optimal_theta_res_loop["x"] = optimal_coop_theta
                            if optimal_theta_res_loop["fun"] < optimal_theta_res["fun"]:
                                optimal_theta_res = optimal_theta_res_loop
                        optimal_theta = optimal_theta_res["x"]
                    elif (
                        self.options["hyper_opt"] == "TNC"
                        and self.options["n_start"] >= 1
                    ):
                        theta_all_loops = 10**theta_all_loops
                        for theta0_loop in theta_all_loops:
                            optimal_theta_res_loop = optimize.minimize(
                                minus_cooperative_reduced_likelihood_function,  # adjusted likelihood function
                                theta0_loop[
                                    self.active_comp_var
                                ],  # only optimize part of theta
                                method="TNC",
                                jac=grad_minus_cooperative_reduced_likelihood_function,
                                bounds=bounds_hyp,
                                options={"maxiter": 100},
                            )
                            if self.name not in ["MGP"]:
                                optimal_coop_theta = np.log10(self.coop_theta)
                            else:
                                optimal_coop_theta = self.coop_theta
                            optimal_coop_theta[self.active_comp_var] = (
                                optimal_theta_res_loop["x"]
                            )
                            optimal_theta_res_loop["x"] = optimal_coop_theta
                            # augment "x" in optimal_theta_res_loop with theta_best of other indices
                            if optimal_theta_res_loop["fun"] < optimal_theta_res["fun"]:
                                optimal_theta_res = optimal_theta_res_loop
                        optimal_theta = optimal_theta_res["x"]
                    else:
                        optimal_theta = theta0

                    if self.name not in ["MGP"]:
                        optimal_theta = 10**optimal_theta
                    optimal_rlf_value, optimal_par = self._reduced_likelihood_function(
                        theta=optimal_theta
                    )

                    # Compare the new optimizer to the best previous one
                    if k > 0:
                        if np.isinf(optimal_rlf_value):
                            raise RuntimeError(
                                "Cannot train the model: infinite likelihood found"
                            )
                        else:
                            if optimal_rlf_value >= self.best_iteration_fail:
                                if optimal_rlf_value > best_optimal_rlf_value:
                                    best_optimal_rlf_value = optimal_rlf_value
                                    best_optimal_par = optimal_par
                                    best_optimal_theta = optimal_theta
                                else:
                                    if (
                                        self.best_iteration_fail
                                        > best_optimal_rlf_value
                                    ):
                                        best_optimal_theta = self._thetaMemory
                                        (
                                            best_optimal_rlf_value,
                                            best_optimal_par,
                                        ) = self._reduced_likelihood_function(
                                            theta=best_optimal_theta
                                        )
                    else:
                        if np.isinf(optimal_rlf_value):
                            stop += 1
                        else:
                            best_optimal_rlf_value = optimal_rlf_value
                            best_optimal_par = optimal_par
                            best_optimal_theta = optimal_theta
                    k += 1
                except ValueError as ve:
                    # raise ve
                    # If iteration is max when fmin_cobyla fail is not reached
                    if self.retry > 0:
                        self.retry -= 1
                        k += 1
                        stop += 1
                        # One evaluation objectif function is done at least
                        if self.best_iteration_fail is not None:
                            if self.best_iteration_fail > best_optimal_rlf_value:
                                best_optimal_theta = self._thetaMemory
                                (
                                    best_optimal_rlf_value,
                                    best_optimal_par,
                                ) = self._reduced_likelihood_function(
                                    theta=best_optimal_theta
                                )
                    # Optimization fail
                    elif best_optimal_par == []:
                        print("Optimization failed. Try increasing the ``nugget``")
                        raise ve
                    # Break the while loop
                    else:
                        k = stop + 1
                        print("fmin_cobyla failed but the best value is retained")

            if "KPLSK" in self.name:
                if self.options["eval_noise"]:
                    # best_optimal_theta contains [theta, noise] if eval_noise = True
                    theta = best_optimal_theta[:-1]
                else:
                    # best_optimal_theta contains [theta] if eval_noise = False
                    theta = best_optimal_theta

                if exit_function:
                    return best_optimal_rlf_value, best_optimal_par, best_optimal_theta

                if self.options["corr"] == "squar_exp":
                    self.options["theta0"] = (theta * self.coeff_pls**2).sum(1)
                else:
                    self.options["theta0"] = (theta * np.abs(self.coeff_pls)).sum(1)

                self.options["n_comp"] = int(self.nx)
                limit = 10 * self.options["n_comp"]
                self.best_iteration_fail = None
                exit_function = True
        return best_optimal_rlf_value, best_optimal_par, best_optimal_theta

    def _predict_derivatives(self, x, kx):
        """
        Not implemented yet.
        Evaluates the derivatives at a set of points.
        """
        raise NotImplementedError(
            "Derivative prediction is not available for cooperative Kriging."
        )

    def _predict_variance_derivatives(self, x):
        """
        Not implemented yet.
        Provide the derivative of the variance of the model at a set of points.
        """
        raise NotImplementedError(
            "Derivative prediction is not available for cooperative Kriging."
        )



# coding: utf-8
"""
Author: Remi Lafage <remi.lafage@onera.fr>
This package is distributed under New BSD license.
"""

import os
import unittest
from multiprocessing import Pool
from typing import Optional

import numpy as np

from smt.applications import EGO
from smt.applications.ego import Evaluator
from smt.applications.mixed_integer import (
    MixedIntegerContext,
    MixedIntegerSamplingMethod,
)
from smt.problems import Branin, Rosenbrock
from smt.sampling_methods import LHS, FullFactorial
from smt.surrogate_models import (
    GEKPLS,
    GPX,
    KPLS,
    KRG,
    CategoricalVariable,
    DesignSpace,
    FloatVariable,
    IntegerVariable,
    MixIntKernelType,
    OrdinalVariable,
)
from smt.surrogate_models.gpx import GPX_AVAILABLE
from smt.utils.sm_test_case import SMTestCase
from sys import argv

try:
    import matplotlib

    matplotlib.use("Agg")
    NO_MATPLOTLIB = False
except ImportError:
    NO_MATPLOTLIB = True


# This implementation only works with Python > 3.3
class ParallelEvaluator(Evaluator):
    def run(self, fun, x, design_space: Optional = None):
        with Pool(3) as p:
            return np.array(
                [y[0] for y in p.map(fun, [np.atleast_2d(x[i]) for i in range(len(x))])]
            )


class TestEGO(SMTestCase):
    plot = None

    @staticmethod
    def function_test_1d(x):
        # function xsinx
        x = np.reshape(x, (-1,))
        y = np.zeros(x.shape)
        y = (x - 3.5) * np.sin((x - 3.5) / (np.pi))
        return y.reshape((-1, 1))

    def test_evaluator(self):
        x = [[1], [2], [3]]
        expected = TestEGO.function_test_1d(x)
        actual = ParallelEvaluator().run(TestEGO.function_test_1d, x)
        for i in range(len(x)):
            self.assertAlmostEqual(expected[i, 0], actual[i, 0])

    def test_function_test_1d(self):
        n_iter = 15
        xlimits = np.array([[0.0, 25.0]])
        criterion = "EI"
        design_space = DesignSpace(xlimits)

        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            n_doe=3,
            surrogate=KRG(design_space=design_space, print_global=False),
            random_state=42,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=TestEGO.function_test_1d)

        self.assertAlmostEqual(18.9, x_opt.item(), delta=1)
        self.assertAlmostEqual(-15.1, y_opt.item(), delta=1)

    @unittest.skipIf(not GPX_AVAILABLE, "GPX not available")
    def test_function_test_GPX_1d(self):
        n_iter = 15
        xlimits = np.array([[0.0, 25.0]])
        criterion = "EI"
        design_space = DesignSpace(xlimits)
        surrogate = GPX(design_space=design_space)

        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            n_doe=3,
            surrogate=surrogate,
            random_state=42,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=TestEGO.function_test_1d)

        self.assertAlmostEqual(18.9, x_opt.item(), delta=1)
        self.assertAlmostEqual(-15.1, y_opt.item(), delta=1)

    def test_function_ego_noisy_KRG_1d(self):
        n_iter = 15
        xlimits = np.array([[0.0, 25.0]])
        criterion = "EI"
        design_space = DesignSpace(xlimits)
        noise0 = [1e-1]

        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            n_doe=3,
            surrogate=KRG(design_space=design_space, print_global=False, noise0=noise0),
            random_state=42,
            is_ri=True,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=TestEGO.function_test_1d)

        self.assertAlmostEqual(18.9, x_opt.item(), delta=1)
        self.assertAlmostEqual(-15.1, y_opt.item(), delta=1)

    def test_function_test_1d_parallel(self):
        n_iter = 3
        xlimits = np.array([[0.0, 25.0]])
        design_space = DesignSpace(xlimits)

        criterion = "SBO"
        n_parallel = 3

        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            n_doe=3,
            surrogate=KRG(design_space=design_space, print_global=False),
            n_parallel=n_parallel,
            evaluator=ParallelEvaluator(),
            random_state=42,
        )
        x_opt, y_opt, _, _, _ = ego.optimize(fun=TestEGO.function_test_1d)

        self.assertAlmostEqual(18.9, x_opt.item(), delta=1)
        self.assertAlmostEqual(-15.1, y_opt.item(), delta=1)

    def test_EGO_free_vs_noisy(self):
        # Ajouter les points DOE
        xdoe = np.atleast_2d([0, 7, 25]).T

        n_iter = 8  # the number of points one wants to infill to find the minimum
        xlimits = np.array([[0.0, 25.0]])
        random_state = 42  # for reproducibility
        noise0 = [2e-1]
        criterion = "EI"  #'EI' or 'SBO' or 'LCB'
        n_start = 20
        # Model
        design_space = DesignSpace(xlimits, random_state=random_state)
        surrogate = KRG(
            design_space=design_space,
            print_global=False,
            noise0=noise0,
        )
        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=xdoe,
            surrogate=surrogate,
            random_state=random_state,
            n_start=n_start,
        )

        ego_ri = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=xdoe,
            surrogate=surrogate,
            random_state=random_state,
            n_start=n_start,
            is_ri=True,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=self.function_test_1d)
        x_opt_ri, y_opt_ri, _, _, _ = ego_ri.optimize(fun=self.function_test_1d)

        # The optimum found is far from the expected result without re-interpolation
        self.assertNotAlmostEqual(18.9, x_opt.item(), delta=1)
        self.assertNotAlmostEqual(-15.1, y_opt.item(), delta=1)
        # The minimum found here is y =~ 0 instead of -15.1

        # The optimum found is pretty close to the expected result with re-interpolation
        self.assertAlmostEqual(18.9, x_opt_ri.item(), delta=1)
        self.assertAlmostEqual(-15.1, y_opt_ri.item(), delta=1)

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_rosenbrock_2D(self):
        n_iter = 50
        fun = Rosenbrock(ndim=2)
        xlimits = fun.xlimits
        criterion = "LCB"  #'EI' or 'SBO' or 'LCB'
        random_state = 42
        design_space = DesignSpace(xlimits, random_state=random_state)
        xdoe = FullFactorial(xlimits=xlimits)(10)
        ego = EGO(
            n_start=30,
            xdoe=xdoe,
            n_iter=n_iter,
            criterion=criterion,
            surrogate=KRG(design_space=design_space, n_start=25, print_global=False),
            random_state=random_state,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=fun)
        np.testing.assert_allclose([1, 1], x_opt, atol=0.55)
        self.assertAlmostEqual(0.0, y_opt.item(), delta=1)

    def test_rosenbrock_2D_SBO(self):
        n_iter = 10
        fun = Rosenbrock(ndim=2)
        xlimits = fun.xlimits
        criterion = "SBO"  #'EI' or 'SBO' or 'LCB'
        design_space = DesignSpace(xlimits)

        xdoe = FullFactorial(xlimits=xlimits)(50)
        ego = EGO(
            xdoe=xdoe,
            n_iter=n_iter,
            criterion=criterion,
            surrogate=KRG(design_space=design_space, print_global=False),
            random_state=42,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=fun)
        np.testing.assert_allclose([1, 1], x_opt, atol=1)
        self.assertAlmostEqual(0.0, y_opt.item(), delta=1)

    @unittest.skipIf(not GPX_AVAILABLE, "GPX not available")
    def test_rosenbrock_2D_GPX(self):
        n_iter = 10
        fun = Rosenbrock(ndim=2)
        xlimits = fun.xlimits
        criterion = "EI"
        design_space = DesignSpace(xlimits)
        surrogate = GPX(design_space=design_space)

        xdoe = FullFactorial(xlimits=xlimits)(50)
        ego = EGO(
            xdoe=xdoe,
            n_iter=n_iter,
            criterion=criterion,
            surrogate=surrogate,
            random_state=42,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=fun)
        np.testing.assert_allclose([1, 1], x_opt, atol=1)
        self.assertAlmostEqual(0.0, y_opt.item(), delta=1)

    def test_rosenbrock_2D_noisy_KRG(self):
        n_iter = 20
        fun = Rosenbrock(ndim=2)
        xlimits = fun.xlimits
        criterion = "EI"
        design_space = DesignSpace(xlimits)
        noise0 = [1e-1]

        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            n_doe=3,
            surrogate=KRG(design_space=design_space, print_global=False, noise0=noise0),
            random_state=42,
            is_ri=True,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=fun)
        np.testing.assert_allclose([1, 1], x_opt, atol=1.5)
        self.assertAlmostEqual(0.0, y_opt.item(), delta=1.5)

    # Comment out broken test on CI ubuntu py3.11, fail without error! code exit 2?
    # @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    # def test_rosenbrock_2D_parallel(self):
    #     n_iter = 20
    #     n_parallel = 5
    #     fun = Rosenbrock(ndim=2)
    #     xlimits = fun.xlimits
    #     criterion = "LCB"  #'EI' or 'SBO' or 'LCB'
    #     random_state = 42
    #     design_space = DesignSpace(xlimits, random_state=random_state)

    #     xdoe = FullFactorial(xlimits=xlimits)(10)
    #     qEI = "KB"
    #     ego = EGO(
    #         xdoe=xdoe,
    #         n_iter=n_iter,
    #         criterion=criterion,
    #         surrogate=KRG(design_space=design_space, print_global=False),
    #         n_parallel=n_parallel,
    #         qEI=qEI,
    #         evaluator=ParallelEvaluator(),
    #         random_state=random_state,
    #     )

    #     x_opt, y_opt, _, _, _ = ego.optimize(fun=fun)
    #     print("Rosenbrock: ", x_opt)
    #     np.testing.assert_allclose([1, 1], x_opt, atol=0.5)
    #     self.assertAlmostEqual(0.0, y_opt.item(), delta=1)

    def test_branin_2D(self):
        n_iter = 20
        fun = Branin(ndim=2)
        criterion = "LCB"  #'EI' or 'SBO' or 'LCB'
        design_space = fun.design_space
        ego = EGO(
            surrogate=KRG(design_space=design_space, print_global=False),
            n_iter=n_iter,
            criterion=criterion,
            n_doe=10,
            random_state=42,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=fun)
        # 3 optimal points possible: [-pi, 12.275], [pi, 2.275], [9.42478, 2.475]
        self.assertTrue(
            np.allclose([[-3.14, 12.275]], x_opt, rtol=0.25)
            or np.allclose([[3.14, 2.275]], x_opt, rtol=0.25)
            or np.allclose([[9.42, 2.475]], x_opt, rtol=0.25)
        )
        self.assertAlmostEqual(0.39, y_opt.item(), delta=0.8)

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_branin_2D_parallel(self):
        n_iter = 10
        fun = Branin(ndim=2)
        n_parallel = 5
        xlimits = fun.xlimits
        criterion = "EI"  #'EI' or 'SBO' or 'LCB'
        design_space = DesignSpace(xlimits)

        xdoe = FullFactorial(xlimits=xlimits)(10)
        ego = EGO(
            xdoe=xdoe,
            n_iter=n_iter,
            criterion=criterion,
            surrogate=KRG(design_space=design_space, print_global=False),
            n_parallel=n_parallel,
            random_state=42,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=fun)

        # 3 optimal points possible: [-pi, 12.275], [pi, 2.275], [9.42478, 2.475]
        self.assertTrue(
            np.allclose([[-3.14, 12.275]], x_opt, rtol=0.5)
            or np.allclose([[3.14, 2.275]], x_opt, rtol=0.5)
            or np.allclose([[9.42, 2.475]], x_opt, rtol=0.5)
        )
        print("Branin=", x_opt)
        self.assertAlmostEqual(0.39, y_opt.item(), delta=1)

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_branin_2D_mixed_parallel(self):
        n_parallel = 5
        n_iter = 20
        fun = Branin(ndim=2)
        xlimits = fun.xlimits
        criterion = "EI"  #'EI' or 'SBO' or 'LCB'
        qEI = "CLmin"
        random_state = 42
        design_space = DesignSpace(
            [
                IntegerVariable(*xlimits[0]),
                FloatVariable(*xlimits[1]),
            ],
            random_state=random_state,
        )
        sm = KRG(design_space=design_space, print_global=False, n_start=25)
        mixint = MixedIntegerContext(design_space)
        sampling = mixint.build_sampling_method()
        xdoe = sampling(10)

        ego = EGO(
            xdoe=xdoe,
            n_iter=n_iter,
            criterion=criterion,
            n_parallel=n_parallel,
            qEI=qEI,
            n_start=30,
            evaluator=ParallelEvaluator(),
            surrogate=sm,
            random_state=random_state,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=fun)
        # 3 optimal points possible: [-pi, 12.275], [pi, 2.275], [9.42478, 2.475]
        self.assertTrue(
            np.allclose([[-3, 12.275]], x_opt, rtol=0.2)
            or np.allclose([[3, 2.275]], x_opt, rtol=0.2)
            or np.allclose([[9, 2.475]], x_opt, rtol=0.2)
        )
        self.assertAlmostEqual(0.494, y_opt.item(), delta=1)

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_branin_2D_mixed(self):
        n_iter = 20
        fun = Branin(ndim=2)
        xlimits = fun.xlimits
        random_state = 42
        design_space = DesignSpace(
            [
                IntegerVariable(*xlimits[0]),
                FloatVariable(*xlimits[1]),
            ],
            random_state=random_state,
        )
        criterion = "EI"  #'EI' or 'SBO' or 'LCB'

        sm = KRG(design_space=design_space, print_global=False)
        sampling = MixedIntegerSamplingMethod(FullFactorial, design_space)
        xdoe = sampling(10)

        ego = EGO(
            xdoe=xdoe,
            n_iter=n_iter,
            criterion=criterion,
            surrogate=sm,
            enable_tunneling=False,
            random_state=random_state,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=fun)
        # 3 optimal points possible: [-pi, 12.275], [pi, 2.275], [9.42478, 2.475]
        self.assertTrue(
            np.allclose([[-3, 12.275]], x_opt, rtol=0.2)
            or np.allclose([[3, 2.275]], x_opt, rtol=0.2)
            or np.allclose([[9, 2.475]], x_opt, rtol=0.2)
        )
        self.assertAlmostEqual(0.494, y_opt.item(), delta=1)

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_branin_2D_mixed_tunnel(self):
        n_iter = 20
        fun = Branin(ndim=2)
        xlimits = fun.xlimits
        random_state = 42
        design_space = DesignSpace(
            [
                IntegerVariable(*xlimits[0]),
                FloatVariable(*xlimits[1]),
            ],
            random_state=random_state,
        )
        criterion = "EI"  #'EI' or 'SBO' or 'LCB'

        sm = KRG(design_space=design_space, print_global=False)
        sampling = MixedIntegerSamplingMethod(FullFactorial, design_space)
        xdoe = sampling(30)

        ego = EGO(
            xdoe=xdoe,
            n_iter=n_iter,
            criterion=criterion,
            surrogate=sm,
            enable_tunneling=True,
            random_state=random_state,
        )

        x_opt, y_opt, _, _, _ = ego.optimize(fun=fun)
        # 3 optimal points possible: [-pi, 12.275], [pi, 2.275], [9.42478, 2.475]
        self.assertTrue(
            np.allclose([[-3, 12.275]], x_opt, rtol=2)
            or np.allclose([[3, 2.275]], x_opt, rtol=2)
            or np.allclose([[9, 2.475]], x_opt, rtol=2)
        )
        self.assertAlmostEqual(0.494, y_opt.item(), delta=2)

    @staticmethod
    def function_test_mixed_integer(X):
        # float
        x1 = X[:, 0]
        #  XType.ENUM 1
        c1 = X[:, 1]
        x2 = c1 == 0
        x3 = c1 == 1
        x4 = c1 == 2
        #  XType.ENUM 2
        c2 = X[:, 2]
        x5 = c2 == 0
        x6 = c2 == 1
        # int
        i = X[:, 3]

        y = (
            (x2 + 2 * x3 + 3 * x4) * x5 * x1
            + (x2 + 2 * x3 + 3 * x4) * x6 * 0.95 * x1
            + i
        )
        return y

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_ego_mixed_integer(self):
        n_iter = 15
        n_doe = 5
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red", "green"]),
                CategoricalVariable(["large", "small"]),
                OrdinalVariable([0, 2, 3]),
            ],
            random_state=42,
        )
        samp = MixedIntegerSamplingMethod(
            LHS, design_space, criterion="ese", random_state=design_space.random_state
        )
        xdoe = samp(n_doe)

        criterion = "EI"  #'EI' or 'SBO' or 'LCB'
        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=xdoe,
            surrogate=KRG(design_space=design_space, print_global=False),
            enable_tunneling=False,
            random_state=design_space.random_state,
        )
        _, y_opt, _, _, _ = ego.optimize(fun=TestEGO.function_test_mixed_integer)

        self.assertAlmostEqual(-15, y_opt.item(), delta=5)

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_ego_mixed_integer_gower_distance(self):
        n_iter = 15
        n_doe = 5
        random_state = 42
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red", "green"]),
                CategoricalVariable(["large", "small"]),
                IntegerVariable(0, 2),
            ],
            random_state=random_state,
        )
        samp = MixedIntegerSamplingMethod(
            LHS, design_space, criterion="ese", random_state=design_space.random_state
        )
        xdoe = samp(n_doe)

        criterion = "EI"  #'EI' or 'SBO' or 'LCB'
        ego = EGO(
            n_start=30,
            n_iter=n_iter,
            criterion=criterion,
            xdoe=xdoe,
            surrogate=KRG(
                n_start=25,
                design_space=design_space,
                categorical_kernel=MixIntKernelType.GOWER,
                print_global=False,
            ),
            enable_tunneling=False,
            random_state=design_space.random_state,
        )
        _, y_opt, _, _, _ = ego.optimize(fun=TestEGO.function_test_mixed_integer)

        self.assertAlmostEqual(-15, y_opt.item(), delta=5)

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_ego_mixed_integer_hierarchical_NN(self):
        random_state = 42

        def f_neu(x1, x2, x3, x4):
            if x4 == 0:
                return 2 * x1 + x2 - 0.5 * x3
            elif x4 == 1:
                return -x1 + 2 * x2 - 0.5 * x3
            elif x4 == 2:
                return -x1 + x2 + 0.5 * x3
            else:
                raise ValueError(f"Unexpected x4: {x4}")

        def f1(x1, x2, x3, x4, x5):
            return f_neu(x1, x2, x3, x4) + x5**2

        def f2(x1, x2, x3, x4, x5, x6):
            return f_neu(x1, x2, x3, x4) + (x5**2) + 0.3 * x6

        def f3(x1, x2, x3, x4, x5, x6, x7):
            return f_neu(x1, x2, x3, x4) + (x5**2) + 0.3 * x6 - 0.1 * x7**3

        def f_hv(X, eval_is_acting):
            y = []
            for i, x in enumerate(X):
                deltai = eval_is_acting[i]
                x3_decoded = design_space.decode_values(x, i_dv=3)[0]
                if np.sum(deltai) == 6:
                    y.append(f1(x[1], x[2], x3_decoded, x[4], x[5]))
                elif np.sum(deltai) == 7:
                    y.append(f2(x[1], x[2], x3_decoded, x[4], x[5], x[6]))
                elif np.sum(deltai) == 8:
                    y.append(f3(x[1], x[2], x3_decoded, x[4], x[5], x[6], x[7]))
                else:
                    raise ValueError(f"Unexpected x0: {x[0]}")
            return np.array(y)

        random_state = 42
        design_space = DesignSpace(
            [
                OrdinalVariable(values=[1, 2, 3]),  # x0
                FloatVariable(-5, 2),
                FloatVariable(-5, 2),
                OrdinalVariable(values=[8, 16, 32, 64, 128, 256]),  # x3
                CategoricalVariable(values=["ReLU", "SELU", "ISRLU"]),  # x4
                IntegerVariable(0, 5),  # x5
                IntegerVariable(0, 5),  # x6
                IntegerVariable(0, 5),  # x7
            ],
            random_state=random_state,
        )

        # x6 is active when x0 >= 2
        design_space.declare_decreed_var(decreed_var=6, meta_var=0, meta_value=[2, 3])
        # x7 is active when x0 >= 3
        design_space.declare_decreed_var(decreed_var=7, meta_var=0, meta_value=3)

        n_doe = 5

        neutral_var_ds = DesignSpace(design_space.design_variables[1:])
        sampling = MixedIntegerSamplingMethod(
            LHS, neutral_var_ds, criterion="ese", random_state=random_state
        )
        x_cont = sampling(3 * n_doe)

        xdoe1 = np.zeros((n_doe, 8))
        x_cont2 = x_cont[:n_doe, :5]
        xdoe1[:, 0] = np.zeros(n_doe)
        xdoe1[:, 1:6] = x_cont2
        # ydoe1 = f_hv(xdoe1)

        xdoe1 = np.zeros((n_doe, 8))
        xdoe1[:, 0] = np.zeros(n_doe)
        xdoe1[:, 1:6] = x_cont2

        xdoe2 = np.zeros((n_doe, 8))
        x_cont2 = x_cont[n_doe : 2 * n_doe, :6]
        xdoe2[:, 0] = np.ones(n_doe)
        xdoe2[:, 1:7] = x_cont2
        # ydoe2 = f_hv(xdoe2)

        xdoe2 = np.zeros((n_doe, 8))
        xdoe2[:, 0] = np.ones(n_doe)
        xdoe2[:, 1:7] = x_cont2

        xdoe3 = np.zeros((n_doe, 8))
        xdoe3[:, 0] = 2 * np.ones(n_doe)
        xdoe3[:, 1:] = x_cont[2 * n_doe :, :]
        # ydoe3 = f_hv(xdoe3)

        Xt = np.concatenate((xdoe1, xdoe2, xdoe3), axis=0)
        # Yt = np.concatenate((ydoe1, ydoe2, ydoe3), axis=0)

        n_iter = 10
        criterion = "EI"

        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=Xt,
            surrogate=KRG(
                design_space=design_space,
                categorical_kernel=MixIntKernelType.HOMO_HSPHERE,
                theta0=[1e-2],
                n_start=5,
                corr="abs_exp",
                print_global=False,
            ),
            enable_tunneling=False,
            random_state=random_state,
        )

        x_opt, y_opt, dnk, x_data, y_data = ego.optimize(fun=f_hv)
        x_corr, eval_is_acting = design_space.correct_get_acting(
            [2, -5, -5, 5, 0, 0, 0, 5]
        )

        self.assertAlmostEqual(
            f_hv(np.atleast_2d(x_corr), eval_is_acting),
            y_opt.item(),
            delta=18,
        )

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_ego_mixed_integer_hierarchical_Goldstein(self):
        def H(x1, x2, x3, x4, z3, z4, x5, cos_term):
            h = (
                53.3108
                + 0.184901 * x1
                - 5.02914 * x1**3 * 10 ** (-6)
                + 7.72522 * x1**z3 * 10 ** (-8)
                - 0.0870775 * x2
                - 0.106959 * x3
                + 7.98772 * x3**z4 * 10 ** (-6)
                + 0.00242482 * x4
                + 1.32851 * x4**3 * 10 ** (-6)
                - 0.00146393 * x1 * x2
                - 0.00301588 * x1 * x3
                - 0.00272291 * x1 * x4
                + 0.0017004 * x2 * x3
                + 0.0038428 * x2 * x4
                - 0.000198969 * x3 * x4
                + 1.86025 * x1 * x2 * x3 * 10 ** (-5)
                - 1.88719 * x1 * x2 * x4 * 10 ** (-6)
                + 2.50923 * x1 * x3 * x4 * 10 ** (-5)
                - 5.62199 * x2 * x3 * x4 * 10 ** (-5)
            )
            if cos_term:
                h += 5.0 * np.cos(2.0 * np.pi * (x5 / 100.0)) - 2.0
            return h

        def f1(x1, x2, z1, z2, z3, z4, x5, cos_term):
            c1 = z2 == 0
            c2 = z2 == 1
            c3 = z2 == 2

            c4 = z3 == 0
            c5 = z3 == 1
            c6 = z3 == 2

            y = (
                c4
                * (
                    c1 * H(x1, x2, 20, 20, z3, z4, x5, cos_term)
                    + c2 * H(x1, x2, 50, 20, z3, z4, x5, cos_term)
                    + c3 * H(x1, x2, 80, 20, z3, z4, x5, cos_term)
                )
                + c5
                * (
                    c1 * H(x1, x2, 20, 50, z3, z4, x5, cos_term)
                    + c2 * H(x1, x2, 50, 50, z3, z4, x5, cos_term)
                    + c3 * H(x1, x2, 80, 50, z3, z4, x5, cos_term)
                )
                + c6
                * (
                    c1 * H(x1, x2, 20, 80, z3, z4, x5, cos_term)
                    + c2 * H(x1, x2, 50, 80, z3, z4, x5, cos_term)
                    + c3 * H(x1, x2, 80, 80, z3, z4, x5, cos_term)
                )
            )
            return y

        def f2(x1, x2, x3, z2, z3, z4, x5, cos_term):
            c1 = z2 == 0
            c2 = z2 == 1
            c3 = z2 == 2

            y = (
                c1 * H(x1, x2, x3, 20, z3, z4, x5, cos_term)
                + c2 * H(x1, x2, x3, 50, z3, z4, x5, cos_term)
                + c3 * H(x1, x2, x3, 80, z3, z4, x5, cos_term)
            )
            return y

        def f3(x1, x2, x4, z1, z3, z4, x5, cos_term):
            c1 = z1 == 0
            c2 = z1 == 1
            c3 = z1 == 2

            y = (
                c1 * H(x1, x2, 20, x4, z3, z4, x5, cos_term)
                + c2 * H(x1, x2, 50, x4, z3, z4, x5, cos_term)
                + c3 * H(x1, x2, 80, x4, z3, z4, x5, cos_term)
            )
            return y

        def f_hv(X, eval_is_acting: Optional = None):
            y = []
            for x in X:
                if x[0] == 0:
                    y.append(
                        f1(x[2], x[3], x[7], x[8], x[9], x[10], x[6], cos_term=x[1])
                    )
                elif x[0] == 1:
                    y.append(
                        f2(x[2], x[3], x[4], x[8], x[9], x[10], x[6], cos_term=x[1])
                    )
                elif x[0] == 2:
                    y.append(
                        f3(x[2], x[3], x[5], x[7], x[9], x[10], x[6], cos_term=x[1])
                    )
                elif x[0] == 3:
                    y.append(
                        H(x[2], x[3], x[4], x[5], x[9], x[10], x[6], cos_term=x[1])
                    )
                else:
                    raise ValueError
            return np.array(y)

        random_state = 0
        ds = DesignSpace(
            [
                CategoricalVariable(values=[0, 1, 2, 3]),  # meta
                OrdinalVariable(values=[0, 1]),  # x1
                FloatVariable(0, 100),  # x2
                FloatVariable(0, 100),
                FloatVariable(0, 100),
                FloatVariable(0, 100),
                FloatVariable(0, 100),
                IntegerVariable(0, 2),  # x7
                IntegerVariable(0, 2),
                IntegerVariable(0, 2),
                IntegerVariable(0, 2),
            ],
            random_state=random_state,
        )

        # x4 is acting if meta == 1, 3
        ds.declare_decreed_var(decreed_var=4, meta_var=0, meta_value=[1, 3])
        # x5 is acting if meta == 2, 3
        ds.declare_decreed_var(decreed_var=5, meta_var=0, meta_value=[2, 3])
        # x7 is acting if meta == 0, 2
        ds.declare_decreed_var(decreed_var=7, meta_var=0, meta_value=[0, 2])
        # x8 is acting if meta == 0, 1
        ds.declare_decreed_var(decreed_var=8, meta_var=0, meta_value=[0, 1])

        n_doe = 25
        samp = MixedIntegerSamplingMethod(
            LHS, ds, criterion="ese", random_state=ds.random_state
        )
        Xt, x_is_active = samp(n_doe, return_is_acting=True)

        n_iter = 10
        criterion = "EI"

        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=Xt,
            surrogate=KRG(
                design_space=ds,
                categorical_kernel=MixIntKernelType.HOMO_HSPHERE,
                theta0=[1e-2],
                n_start=10,
                corr="squar_exp",
                print_global=False,
            ),
            verbose=True,
            enable_tunneling=False,
            random_state=random_state,
            n_start=25,
        )

        x_opt, y_opt, dnk, x_data, y_data = ego.optimize(fun=f_hv)
        self.assertAlmostEqual(
            9.022,
            y_opt.item(),
            delta=25,
        )

    def test_ego_mixed_integer_homo_gaussian(self):
        n_iter = 15
        random_state = 42
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red", "green"]),
                CategoricalVariable(["large", "small"]),
                IntegerVariable(0, 2),
            ],
            random_state=random_state,
        )
        n_doe = 5
        sampling = MixedIntegerSamplingMethod(
            LHS,
            design_space,
            criterion="ese",
            random_state=random_state,
            output_in_folded_space=True,
        )
        xdoe = sampling(n_doe)
        criterion = "EI"  #'EI' or 'SBO' or 'LCB'

        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=xdoe,
            surrogate=KRG(
                design_space=design_space,
                categorical_kernel=MixIntKernelType.EXP_HOMO_HSPHERE,
                print_global=False,
                hyper_opt="Cobyla",
            ),
            enable_tunneling=False,
            random_state=random_state,
        )
        _, y_opt, _, _, _ = ego.optimize(fun=TestEGO.function_test_mixed_integer)

        self.assertAlmostEqual(-15, y_opt.item(), delta=5)

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_ego_mixed_integer_homo_gaussian_pls(self):
        n_iter = 15
        random_state = 42
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red", "green"]),
                CategoricalVariable(["large", "small"]),
                IntegerVariable(0, 2),
            ],
            random_state=random_state,
        )
        sampling = MixedIntegerSamplingMethod(
            LHS,
            design_space,
            criterion="ese",
            random_state=random_state,
            output_in_folded_space=True,
        )
        n_doe = 5
        xdoe = sampling(n_doe)
        criterion = "EI"  #'EI' or 'SBO' or 'LCB'
        sm = KPLS(
            print_global=False,
            design_space=design_space,
            categorical_kernel=MixIntKernelType.EXP_HOMO_HSPHERE,
            n_comp=1,
            cat_kernel_comps=[2, 2],
        )
        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=xdoe,
            surrogate=sm,
            enable_tunneling=False,
            random_state=random_state,
        )
        _, y_opt, _, _, _ = ego.optimize(fun=TestEGO.function_test_mixed_integer)

        self.assertAlmostEqual(-15, y_opt.item(), delta=5)

    def test_ydoe_option(self):
        n_iter = 15
        fun = Branin(ndim=2)
        xlimits = fun.xlimits
        criterion = "LCB"  #'EI' or 'SBO' or 'LCB'
        random_state = 42
        design_space = DesignSpace(xlimits, random_state=random_state)
        xdoe = FullFactorial(xlimits=xlimits)(9)
        ydoe = fun(xdoe)
        ego = EGO(
            xdoe=xdoe,
            ydoe=ydoe,
            n_iter=n_iter,
            criterion=criterion,
            surrogate=KRG(
                design_space=design_space, hyper_opt="Cobyla", print_global=False
            ),
            random_state=random_state,
        )
        _, y_opt, _, _, _ = ego.optimize(fun=fun)

        self.assertAlmostEqual(0.39, y_opt.item(), delta=1)

    def test_find_best_point(self):
        fun = TestEGO.function_test_1d
        xlimits = np.array([[0.0, 25.0]])
        random_state = 42
        design_space = DesignSpace(xlimits, random_state=random_state)
        xdoe = FullFactorial(xlimits=xlimits)(3)
        ydoe = fun(xdoe)
        ego = EGO(
            xdoe=xdoe,
            ydoe=ydoe,
            n_iter=1,
            criterion="LCB",
            surrogate=KRG(design_space=design_space, print_global=False),
            n_start=30,
            enable_tunneling=False,
            random_state=random_state,
        )
        _, _, _, _, _ = ego.optimize(fun=fun)
        x, _ = ego._find_best_point(xdoe, ydoe, enable_tunneling=False)
        self.assertAlmostEqual(6.5, x.item(), delta=1)

    @staticmethod
    def initialize_ego_gek(func="exp", criterion="LCB"):
        from smt.problems import TensorProduct

        class TensorProductIndirect(TensorProduct):
            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                self.super = super()

            def _evaluate(self, x, kx):
                assert kx is None
                response = self.super._evaluate(x, kx)
                sens = np.hstack(
                    [self.super._evaluate(x, ki) for ki in range(x.shape[1])]
                )
                return np.hstack((response, sens))

        fun = TensorProductIndirect(ndim=2, func=func)
        random_state = 42
        design_space = DesignSpace(fun.xlimits, random_state=42)

        # Construction of the DOE
        sampling = LHS(xlimits=fun.xlimits, criterion="m", random_state=random_state)
        xdoe = sampling(20)
        ydoe = fun(xdoe)

        # Build the GEKPLS surrogate model
        n_comp = 2
        sm = GEKPLS(
            theta0=[1e-2] * n_comp,
            design_space=design_space,
            extra_points=1,
            eval_comp_treshold=0.8,
            print_prediction=False,
            n_comp=n_comp,
        )

        # Build the EGO optimizer and optimize
        ego = EGO(
            xdoe=xdoe,
            ydoe=ydoe,
            n_iter=5,
            criterion=criterion,
            surrogate=sm,
            n_start=30,
            enable_tunneling=False,
            random_state=random_state,
        )

        return ego, fun

    def test_ego_random_stateing(self):
        def f_obj(X):
            """
            s01 objective

            Parameters
            ----------
            point: array_like
                point to evaluate
            """
            PI = 3.14159265358979323846
            x = X[:, 0]
            # categorial variable
            c = X[:, 1]
            x = np.abs(x)
            c1 = c == 0
            c2 = c == 1
            c3 = c == 2
            c4 = c == 3
            c5 = c == 4
            c6 = c == 5
            c7 = c == 6
            c8 = c == 7
            c9 = c == 8
            c10 = c == 9
            if np.size(c1) == (
                np.sum(c1)
                + np.sum(c2)
                + np.sum(c3)
                + np.sum(c4)
                + np.sum(c5)
                + np.sum(c6)
                + np.sum(c7)
                + np.sum(c8)
                + np.sum(c9)
                + np.sum(c10)
            ):
                y = (
                    c1 * (np.cos(3.6 * PI * (x - 2)) + x - 1)
                    + c2 * (2 * np.cos(1.1 * PI * np.exp(x)) - x / 2 + 2)
                    + c3 * (np.cos(2 * PI * x) + x / 2)
                    + c4 * (x * (np.cos(3.4 * PI * (x - 1)) - (x - 1) / 2))
                    + c5 * (-0.5 * x * x)
                    + c6
                    * (
                        2 * np.power(np.cos(0.25 * PI * np.exp(-np.power(x, 4))), 2)
                        - x / 2
                        + 1
                    )
                    + c7 * (x * (np.cos(3.4 * PI * x)) - x / 2 + 1)
                    + c8 * (x * (-np.cos(7 * 0.5 * PI * x) - x / 2) + 2)
                    + c9 * (-np.power(x, 5) * 0.5 + 1)
                    + c10
                    * (
                        -np.power(np.cos(5 * PI * 0.5 * x), 2) * np.sqrt(x)
                        - 0.5 * np.log(x + 0.5)
                        - 1.3
                    )
                )
            else:
                print("type error")
                print(X)
            return y

        # To define the variables x^{quant} and x^{cat}
        design_space = DesignSpace(
            [
                FloatVariable(0, 1),  # real
                CategoricalVariable(
                    ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
                ),  # 10 possible choices
            ]
        )

        # To define the initial DOE
        random_state = 42  # random_state value for the sampling
        n_doe = 5  # initial doe size
        sampling = MixedIntegerSamplingMethod(
            LHS, design_space, criterion="ese", random_state=random_state
        )
        Xt = sampling(n_doe)
        self.assertAlmostEqual(np.sum(Xt), 33.56885202767958, delta=1e-4)
        Xt = np.array(
            [
                [0.37454012, 1.0],
                [0.95071431, 0.0],
                [0.73199394, 8.0],
                [0.59865848, 6.0],
                [0.15601864, 7.0],
            ]
        )
        # To start the Bayesion optimization
        n_iter = 2  # number of iterations
        criterion = "LCB"  # infill criterion
        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=Xt,
            surrogate=KRG(
                design_space=design_space,
                categorical_kernel=MixIntKernelType.GOWER,
                theta0=[1e-2],
                n_start=25,
                corr="squar_exp",
                hyper_opt="Cobyla",
                print_global=False,
            ),
            verbose=False,
            enable_tunneling=False,
            random_state=random_state,
            n_start=25,
        )
        x_opt, y_opt, dnk, x_data, y_data = ego.optimize(fun=f_obj)
        self.assertAlmostEqual(np.sum(y_data), 8.846225742003778, delta=1e-4)
        self.assertAlmostEqual(np.sum(x_data), 41.81192549000013, delta=1e-4)

    def test_ego_gek(self):
        ego, fun = self.initialize_ego_gek()
        x_opt, _, _, _, _ = ego.optimize(fun=fun)

        self.assertAlmostEqual(-1.0, float(x_opt[0]), delta=1e-4)
        self.assertAlmostEqual(-1.0, float(x_opt[1]), delta=1e-4)

    def test_ei_gek(self):
        ego, fun = self.initialize_ego_gek(func="cos", criterion="EI")
        x_data, y_data = ego._setup_optimizer(fun)
        ego._train_gpr(x_data, y_data)

        # Test the EI value at the following point
        ei = ego.EI(
            np.array(
                [[0.8398599985874058, -0.3240337426231973], [-0.45961638, 0.40808533]]
            )
        )

        self.assertTrue(
            np.allclose(
                ei,
                [
                    [6.83719886e-12, 8.13390235e-02, 9.26624101e-02],
                    [0.00000000e00, 0.00000000e00, 0.00000000e00],
                ],
                atol=1e-2,
            )
        )

    def test_qei_criterion_default(self):
        fun = TestEGO.function_test_1d
        xlimits = np.array([[0.0, 25.0]])
        random_state = 42
        design_space = DesignSpace(xlimits, random_state=random_state)
        xdoe = FullFactorial(xlimits=xlimits)(3)
        ydoe = fun(xdoe)
        ego = EGO(
            xdoe=xdoe,
            ydoe=ydoe,
            n_iter=1,
            n_parallel=2,
            criterion="SBO",
            surrogate=KRG(design_space=design_space, print_global=False),
            n_start=30,
            random_state=random_state,
        )
        ego._setup_optimizer(fun)
        ego.gpr.set_training_values(xdoe, ydoe)
        ego.gpr.train()
        xtest = np.array([[10.0]])
        # test that default virtual point should be equal to 3sigma lower bound kriging interval
        expected = (
            ego.gpr.predict_values(xtest)
            - 3 * np.sqrt(ego.gpr.predict_variances(xtest))
        ).item()
        actual = ego._get_virtual_point(xtest, fun(xtest))[0].item()
        self.assertAlmostEqual(expected, actual)

    @unittest.skipIf(
        int(os.getenv("RUN_SLOW_TESTS", 0)) < 2 or NO_MATPLOTLIB,
        "too slow or matplotlib not installed",
    )
    def test_examples(self):
        self.run_ego_example()
        self.run_ego_parallel_example()
        self.run_ego_mixed_integer_example()

    @staticmethod
    def run_ego_example():
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications import EGO
        from smt.design_space import DesignSpace
        from smt.surrogate_models import KRG

        def function_test_1d(x):
            # function xsinx
            import numpy as np

            x = np.reshape(x, (-1,))
            y = np.zeros(x.shape)
            y = (x - 3.5) * np.sin((x - 3.5) / (np.pi))
            return y.reshape((-1, 1))

        n_iter = 6
        xlimits = np.array([[0.0, 25.0]])

        random_state = 42  # for reproducibility
        design_space = DesignSpace(xlimits, random_state=random_state)
        xdoe = np.atleast_2d([0, 7, 25]).T
        n_doe = xdoe.size

        criterion = "EI"  #'EI' or 'SBO' or 'LCB'

        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=xdoe,
            surrogate=KRG(design_space=design_space, print_global=False),
            random_state=random_state,
        )

        x_opt, y_opt, _, x_data, y_data = ego.optimize(fun=function_test_1d)
        print("Minimum in x={:.1f} with f(x)={:.1f}".format(x_opt.item(), y_opt.item()))

        x_plot = np.atleast_2d(np.linspace(0, 25, 100)).T
        y_plot = function_test_1d(x_plot)

        fig = plt.figure(figsize=[10, 10])
        for i in range(n_iter):
            k = n_doe + i
            x_data_k = x_data[0:k]
            y_data_k = y_data[0:k]
            ego.gpr.set_training_values(x_data_k, y_data_k)
            ego.gpr.train()

            y_gp_plot = ego.gpr.predict_values(x_plot)
            y_gp_plot_var = ego.gpr.predict_variances(x_plot)
            y_ei_plot = -ego.EI(x_plot)

            ax = fig.add_subplot((n_iter + 1) // 2, 2, i + 1)
            ax1 = ax.twinx()
            (ei,) = ax1.plot(x_plot, y_ei_plot, color="red")

            (true_fun,) = ax.plot(x_plot, y_plot)
            (data,) = ax.plot(
                x_data_k, y_data_k, linestyle="", marker="o", color="orange"
            )
            if i < n_iter - 1:
                (opt,) = ax.plot(
                    x_data[k], y_data[k], linestyle="", marker="*", color="r"
                )
            (gp,) = ax.plot(x_plot, y_gp_plot, linestyle="--", color="g")
            sig_plus = y_gp_plot + 3 * np.sqrt(y_gp_plot_var)
            sig_moins = y_gp_plot - 3 * np.sqrt(y_gp_plot_var)
            un_gp = ax.fill_between(
                x_plot.T[0], sig_plus.T[0], sig_moins.T[0], alpha=0.3, color="g"
            )
            lines = [true_fun, data, gp, un_gp, opt, ei]
            fig.suptitle("EGO optimization of $f(x) = x \\sin{x}$")
            fig.subplots_adjust(hspace=0.4, wspace=0.4, top=0.8)
            ax.set_title("iteration {}".format(i + 1))
            fig.legend(
                lines,
                [
                    "f(x)=xsin(x)",
                    "Given data points",
                    "Kriging prediction",
                    "Kriging 99% confidence interval",
                    "Next point to evaluate",
                    "Expected improvment function",
                ],
            )
        plt.show()
        # Check the optimal point is x_opt=18.9, y_opt =-15.1

    @staticmethod
    def run_ego_mixed_integer_example():
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications import EGO
        from smt.applications.mixed_integer import MixedIntegerContext
        from smt.design_space import (
            CategoricalVariable,
            DesignSpace,
            FloatVariable,
            IntegerVariable,
        )
        from smt.surrogate_models import KRG, MixIntKernelType

        # Regarding the interface, the function to be optimized should handle
        # categorical values as index values in the enumeration type specification.
        # For instance, here "blue" will be passed to the function as the index value 2.
        # This allows to keep the numpy ndarray X handling numerical values.
        def function_test_mixed_integer(X):
            # float
            x1 = X[:, 0]
            #  enum 1
            c1 = X[:, 1]
            x2 = c1 == 0
            x3 = c1 == 1
            x4 = c1 == 2
            #  enum 2
            c2 = X[:, 2]
            x5 = c2 == 0
            x6 = c2 == 1
            # int
            i = X[:, 3]

            y = (
                (x2 + 2 * x3 + 3 * x4) * x5 * x1
                + (x2 + 2 * x3 + 3 * x4) * x6 * 0.95 * x1
                + i
            )
            return y.reshape((-1, 1))

        n_iter = 15
        random_state = 42
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red", "green"]),
                CategoricalVariable(["square", "circle"]),
                IntegerVariable(0, 2),
            ],
            random_state=random_state,
        )

        criterion = "EI"  #'EI' or 'SBO' or 'LCB'
        qEI = "KBRand"
        sm = KRG(
            design_space=design_space,
            categorical_kernel=MixIntKernelType.GOWER,
            hyper_opt="Cobyla",
            print_global=False,
        )
        mixint = MixedIntegerContext(design_space)
        n_doe = 3
        sampling = mixint.build_sampling_method(random_state=random_state)
        xdoe = sampling(n_doe)
        ydoe = function_test_mixed_integer(xdoe)

        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=xdoe,
            ydoe=ydoe,
            surrogate=sm,
            qEI=qEI,
            n_parallel=2,
            random_state=random_state,
        )

        x_opt, y_opt, _, _, y_data = ego.optimize(fun=function_test_mixed_integer)
        print("Minimum in x={} with f(x)={:.1f}".format(x_opt, y_opt.item()))
        # print("Minimum in typed x={}".format(ego.mixint.cast_to_mixed_integer(x_opt)))

        min_ref = -15
        mini = np.zeros(n_iter)
        for k in range(n_iter):
            mini[k] = np.log(np.abs(np.min(y_data[0 : k + n_doe - 1]) - min_ref))
        x_plot = np.linspace(1, n_iter + 0.5, n_iter)
        up = max(np.floor(max(mini)) + 1, -100)
        lo = max(np.floor(min(mini)) - 0.2, -10)
        fig = plt.figure()
        axes = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        axes.plot(x_plot, mini, color="r")
        axes.set_ylim([lo, up])
        plt.title("minimum convergence plot", loc="center")
        plt.xlabel("number of iterations")
        plt.ylabel("log of the difference w.r.t the best")
        plt.show()

    @staticmethod
    def run_ego_parallel_example():
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications import EGO
        from smt.applications.ego import Evaluator
        from smt.surrogate_models import KRG, DesignSpace
        from typing import Optional

        def function_test_1d(x):
            # function xsinx
            import numpy as np

            x = np.reshape(x, (-1,))
            y = np.zeros(x.shape)
            y = (x - 3.5) * np.sin((x - 3.5) / (np.pi))
            return y.reshape((-1, 1))

        n_iter = 3
        n_parallel = 3
        n_start = 50
        xlimits = np.array([[0.0, 25.0]])

        random_state = 42
        design_space = DesignSpace(xlimits, random_state=random_state)
        xdoe = np.atleast_2d([0, 7, 25]).T
        n_doe = xdoe.size

        class ParallelEvaluator(Evaluator):
            """
            Implement Evaluator interface using multiprocessing ThreadPool object (Python 3 only).
            """

            def run(self, fun, x, design_space: Optional = None):
                n_thread = 5
                # Caveat: import are made here due to SMT documentation building process
                from multiprocessing.pool import ThreadPool
                from sys import version_info

                import numpy as np

                if version_info.major == 2:
                    return fun(x)
                # Python 3 only
                with ThreadPool(n_thread) as p:
                    return np.array(
                        [
                            y[0]
                            for y in p.map(
                                fun, [np.atleast_2d(x[i]) for i in range(len(x))]
                            )
                        ]
                    )

        criterion = "EI"  #'EI' or 'SBO' or 'LCB'
        qEI = "KBUB"  # "KB", "KBLB", "KBUB", "KBRand"
        ego = EGO(
            n_iter=n_iter,
            criterion=criterion,
            xdoe=xdoe,
            surrogate=KRG(design_space=design_space, print_global=False),
            n_parallel=n_parallel,
            qEI=qEI,
            n_start=n_start,
            evaluator=ParallelEvaluator(),
            random_state=random_state,
        )

        x_opt, y_opt, _, x_data, y_data = ego.optimize(fun=function_test_1d)
        print("Minimum in x={:.1f} with f(x)={:.1f}".format(x_opt.item(), y_opt.item()))

        x_plot = np.atleast_2d(np.linspace(0, 25, 100)).T
        y_plot = function_test_1d(x_plot)

        fig = plt.figure(figsize=[10, 10])
        for i in range(n_iter):
            k = n_doe + (i) * (n_parallel)
            x_data_k = x_data[0:k]
            y_data_k = y_data[0:k]
            x_data_sub = x_data_k.copy()
            y_data_sub = y_data_k.copy()
            for p in range(n_parallel):
                ego.gpr.set_training_values(x_data_sub, y_data_sub)
                ego.gpr.train()

                y_ei_plot = -ego.EI(x_plot)
                y_gp_plot = ego.gpr.predict_values(x_plot)
                y_gp_plot_var = ego.gpr.predict_variances(x_plot)

                x_data_sub = np.append(x_data_sub, x_data[k + p])
                y_KB = ego._get_virtual_point(np.atleast_2d(x_data[k + p]), y_data_sub)

                y_data_sub = np.append(y_data_sub, y_KB)

                ax = fig.add_subplot(n_iter, n_parallel, i * (n_parallel) + p + 1)
                ax1 = ax.twinx()
                (ei,) = ax1.plot(x_plot, y_ei_plot, color="red")

                (true_fun,) = ax.plot(x_plot, y_plot)
                (data,) = ax.plot(
                    x_data_sub[: -1 - p],
                    y_data_sub[: -1 - p],
                    linestyle="",
                    marker="o",
                    color="orange",
                )
                (virt_data,) = ax.plot(
                    x_data_sub[-p - 1 : -1],
                    y_data_sub[-p - 1 : -1],
                    linestyle="",
                    marker="o",
                    color="g",
                )

                (opt,) = ax.plot(
                    x_data_sub[-1], y_data_sub[-1], linestyle="", marker="*", color="r"
                )
                (gp,) = ax.plot(x_plot, y_gp_plot, linestyle="--", color="g")
                sig_plus = y_gp_plot + 3.0 * np.sqrt(y_gp_plot_var)
                sig_moins = y_gp_plot - 3.0 * np.sqrt(y_gp_plot_var)
                un_gp = ax.fill_between(
                    x_plot.T[0], sig_plus.T[0], sig_moins.T[0], alpha=0.3, color="g"
                )
                lines = [true_fun, data, gp, un_gp, opt, ei, virt_data]
                fig.suptitle("EGOp optimization of $f(x) = x \\sin{x}$")
                fig.subplots_adjust(hspace=0.4, wspace=0.4, top=0.8)
                ax.set_title("iteration {}.{}".format(i, p))
                fig.legend(
                    lines,
                    [
                        "f(x)=xsin(x)",
                        "Given data points",
                        "Kriging prediction",
                        "Kriging 99% confidence interval",
                        "Next point to evaluate",
                        "Expected improvment function",
                        "Virtula data points",
                    ],
                )
        plt.show()


if __name__ == "__main__":
    if "--plot" in argv:
        TestEGO.plot = True
        argv.remove("--plot")
    if "--example" in argv:
        TestEGO.run_ego_mixed_integer_example()
        exit()
    unittest.main()



__version__ = "2.9.4"

__all__ = [
    "surrogate_models",
    "kernels",
    "design_space",
    "applications",
    "examples",
    "sampling_methods",
    "utils",
    "tests",
    "src",
    "problems",
]



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

Random sampling.
"""

import warnings

import numpy as np

from smt.sampling_methods.sampling_method import ScaledSamplingMethod

# Check NumPy version
numpy_version = tuple(
    map(int, np.__version__.split(".")[:2])
)  # Extract major and minor version


class Random(ScaledSamplingMethod):
    def _initialize(self, **kwargs):
        self.options.declare(
            "random_state",
            types=(type(None), int, np.random.RandomState, np.random.Generator),
            desc="Numpy RandomState or Generator object or seed number which controls random draws",
        )

        # Update options values passed by the user here to get 'random_state' option
        self.options.update(kwargs)

        # RandomState and Generator are and have to be initialized once at constructor time,
        # not in _compute to avoid yielding the same dataset again and again
        if numpy_version < (2, 0):  # Version is below 2.0.0
            if isinstance(self.options["random_state"], np.random.RandomState):
                self.random_state = self.options["random_state"]
            elif isinstance(self.options["random_state"], np.random.Generator):
                self.random_state = np.random.RandomState()
                warnings.warn(
                    "numpy.random.Generator initialization of random_state is not implemented for numpy "
                    "versions < 2.0.0. Using the default np.random.RandomState() as random_state. "
                    "Please consider upgrading to numpy version > 2.0.0, or use the legacy numpy.random.RandomState "
                    "class in the future.",
                    FutureWarning,
                )
            elif isinstance(self.options["random_state"], int):
                self.random_state = np.random.RandomState(self.options["random_state"])
            else:
                self.random_state = np.random.RandomState()
        else:
            # Construct a new Generator with the default BitGenerator (PCG64).
            # If passed a Generator, it will be returned unaltered. When passed a legacy
            # RandomState instance it will be coerced to a Generator.
            self.random_state = np.random.default_rng(seed=self.options["random_state"])

    def _compute(self, nt):
        """
        Implemented by sampling methods to compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the unit hypercube.
        """
        xlimits = self.options["xlimits"]
        nx = xlimits.shape[0]
        if numpy_version < (2, 0):  # Version is below 2.0.0
            return self.random_state.rand(nt, nx)
        else:
            # Create a Generator object with a specified seed (numpy.random_state.rand(nt, nx)
            # is being deprecated)
            return self.random_state.random((nt, nx))
